import datetime
import pandas as pd
import pytest

from thermostat.importers import (
        from_csv,
        normalize_utc_offset,
        )

from thermostat.util.testing import get_data_path

from .fixtures.single_stage import (
        thermostat_bad_equipment_type,
        thermostat_missing_days,
        thermostat_missing_hours,
        thermostat_missing_over_18_days_temperature,
        thermostat_missing_temperature,
        thermostat_type_1,
        thermostat_type_1_cache,
        thermostat_type_1_climate_zone_not_found,
        thermostat_type_1_data_missing_header,
        thermostat_type_1_data_out_of_order,
        thermostat_type_1_metadata_missing_header,
        thermostat_type_1_station_not_found,
        thermostat_type_1_too_many_minutes,
        thermostat_type_1_utc,
        thermostat_type_1_utc_bad,
        thermostat_type_1_utc_bad,
        thermostat_type_1_zip_bad,
        thermostat_type_1_zip_code_not_found,
        )

def test_import_csv(thermostat_type_1):

    def assert_is_series_with_shape(series, shape):
        assert isinstance(series, pd.Series)
        assert series.shape == shape

    assert_is_series_with_shape(thermostat_type_1.cool_runtime_daily, (1461,))
    assert_is_series_with_shape(thermostat_type_1.heat_runtime_daily, (1461,))

    assert_is_series_with_shape(thermostat_type_1.cool_runtime_hourly, (35064,))
    assert_is_series_with_shape(thermostat_type_1.heat_runtime_hourly, (35064,))

    assert_is_series_with_shape(thermostat_type_1.auxiliary_heat_runtime, (35064,))
    assert_is_series_with_shape(thermostat_type_1.emergency_heat_runtime, (35064,))

    assert_is_series_with_shape(thermostat_type_1.temperature_in, (35064,))
    assert_is_series_with_shape(thermostat_type_1.temperature_out, (35064,))

def test_import_csv_cache(thermostat_type_1_cache):
    assert thermostat_type_1_cache is not None

def test_utc_offset(thermostat_type_1_utc, thermostat_type_1_utc_bad):
    assert(normalize_utc_offset("+0") == datetime.timedelta(0))
    assert(normalize_utc_offset("-0") == datetime.timedelta(0))
    assert(normalize_utc_offset("0") == datetime.timedelta(0))
    assert(normalize_utc_offset(0) == datetime.timedelta(0))
    assert(normalize_utc_offset("+6") == datetime.timedelta(0, 21600))
    assert(normalize_utc_offset("-6") == datetime.timedelta(-1, 64800))
    assert(normalize_utc_offset("-0600") == datetime.timedelta(-1, 64800))
    assert(normalize_utc_offset(-6) == datetime.timedelta(-1, 64800))

    with pytest.raises(TypeError) as excinfo:
        normalize_utc_offset('+O')
    assert "Invalid UTC" in str(excinfo)

    with pytest.raises(TypeError) as excinfo:
        normalize_utc_offset('6')
    assert "Invalid UTC" in str(excinfo)

    # Load a thermostat with utc offset == 0
    assert(isinstance(thermostat_type_1_utc.cool_runtime_daily, pd.Series))
    assert(isinstance(thermostat_type_1_utc.cool_runtime_hourly, pd.Series))
    assert(thermostat_type_1_utc_bad is None)

def test_too_many_minutes(thermostat_type_1_too_many_minutes):
    # None of the thermostats in this list should import
    assert len(thermostat_type_1_too_many_minutes) == 0

def test_missing_runtime_hours(thermostat_missing_hours):
    missing_hours = thermostat_missing_hours.heat_runtime_hourly[thermostat_missing_hours.heat_runtime_hourly.isna()]
    missing_hours_days = set(missing_hours.index.date)
    missing_daily = thermostat_missing_hours.heat_runtime_daily[thermostat_missing_hours.heat_runtime_daily.isna()]
    missing_daily_days = set(missing_daily.index.date)
    assert(len(thermostat_missing_hours.heat_runtime_hourly.loc[['2011-05-08 18:00:00']].dropna()) == 0)
    assert(len(thermostat_missing_hours.heat_runtime_daily.loc[['2011-05-08']].dropna()) == 1)
    assert(len(thermostat_missing_hours.heat_runtime_hourly.loc[['2011-05-09 7:00:00']].dropna()) == 0)
    assert(len(thermostat_missing_hours.heat_runtime_hourly.loc[['2011-05-09 8:00:00']].dropna()) == 0)
    assert(len(thermostat_missing_hours.heat_runtime_daily.loc[['2011-05-09']].dropna()) == 1)
    assert(len(thermostat_missing_hours.heat_runtime_daily.loc[['2011-01-01']].dropna()) == 0)
    assert missing_daily_days.difference(missing_hours_days) == set()

def test_missing_temperature_over_18_days(thermostat_missing_over_18_days_temperature):
    # Show that thermostats with temperature_in data of over 18 days missing are still created
    assert len(thermostat_missing_over_18_days_temperature.enough_temp_in[thermostat_missing_over_18_days_temperature.enough_temp_in]) < 347

def test_missing_temperature_hours(thermostat_missing_temperature):
    assert(len(thermostat_missing_temperature.temperature_in.loc[['2011-06-02 6:00:00']].dropna()) == 0)
    assert(len(thermostat_missing_temperature.temperature_in.loc[['2011-06-02 7:00:00']].dropna()) == 0)
    assert(len(thermostat_missing_temperature.temperature_in.loc[['2011-06-02 8:00:00']].dropna()) == 0)
    assert(len(thermostat_missing_temperature.temperature_in.loc[['2012-06-02 18:00:00']].dropna()) == 1)
    assert(len(thermostat_missing_temperature.temperature_in.loc[['2011-06-28 18:00:00']].dropna()) == 1)
    assert(len(thermostat_missing_temperature.temperature_in.loc[['2012-06-29 18:00:00']].dropna()) == 1)
    assert(len(thermostat_missing_temperature.temperature_in.loc[['2014-12-15 20:00:00']].dropna()) == 1)
    assert(len(thermostat_missing_temperature.temperature_in.loc[['2014-12-31 20:00:00']].dropna()) == 1)

    assert thermostat_missing_temperature.enough_temp_in['2011-06-02'] == False
    assert thermostat_missing_temperature.enough_temp_in['2011-06-25'] == True
    assert thermostat_missing_temperature.enough_temp_in['2011-06-26'] == False
    assert thermostat_missing_temperature.enough_temp_in['2011-06-27'] == True
    assert thermostat_missing_temperature.enough_temp_in['2011-06-28'] == True
    assert thermostat_missing_temperature.enough_temp_in['2011-06-29'] == True
    assert thermostat_missing_temperature.enough_temp_in['2011-06-30'] == True
    assert thermostat_missing_temperature.enough_temp_in['2012-08-12'] == False
    assert thermostat_missing_temperature.enough_temp_in['2014-05-28'] == False
    assert thermostat_missing_temperature.enough_temp_in['2014-12-15'] == True
    assert thermostat_missing_temperature.enough_temp_in['2014-12-31'] == True

    assert(len(thermostat_missing_temperature.temperature_out.loc[['2011-06-26 18:00:00']].dropna()) == 0)
    assert(len(thermostat_missing_temperature.temperature_out.loc[['2012-06-26 18:00:00']].dropna()) == 1)
    assert thermostat_missing_temperature.enough_temp_out['2011-06-26'] == False
    assert thermostat_missing_temperature.enough_temp_out['2012-06-26'] == True

def test_missing_days(thermostat_missing_days):
    # Checking what happens
    assert len(thermostat_missing_days) == 0

def test_bad_zipcode(thermostat_type_1_zip_bad):
    assert len(thermostat_type_1_zip_bad) == 0

def test_zip_code_not_found(thermostat_type_1_zip_code_not_found):
    assert len(thermostat_type_1_zip_code_not_found) == 0

def test_station_not_found(thermostat_type_1_station_not_found):
    assert len(thermostat_type_1_station_not_found) == 0

def test_climate_zone_not_found(thermostat_type_1_climate_zone_not_found):
    assert len(thermostat_type_1_climate_zone_not_found) == 0

def test_data_out_of_order(thermostat_type_1_data_out_of_order):
    assert len(thermostat_type_1_data_out_of_order) == 0

def test_data_missing_headers(thermostat_type_1_data_missing_header):
    assert len(thermostat_type_1_data_missing_header) == 0

def test_data_missing_headers(thermostat_type_1_data_missing_header):
    assert len(thermostat_type_1_data_missing_header) == 0

def test_data_bad_equipment_type(thermostat_bad_equipment_type):
    assert len(thermostat_bad_equipment_type) == 0



from thermostat.importers import from_csv
from thermostat.importers import get_single_thermostat
from thermostat.util.testing import get_data_path
from thermostat.core import Thermostat, CoreDaySet
from tempfile import TemporaryDirectory

import pandas as pd
import numpy as np
from numpy import nan

import pytest

''' Abbreviations used in this file:

    XX - Heating system type: fu (furnace), hp (heat pump),  er (electric resistance), ot (other)
    XX - Cooling system type: ce (central), hp (heat pump), 

    YY - backup: eb (electric backup), ne (non electric backup, df (dual fuel), na (N/A)

    Z - speed: 1 (single stage), 2 (two stage), v (variable)
'''

@pytest.fixture(scope="session", params=["../data/two_stage_ert/metadata_furnace_or_boiler_two_stage_central_two_stage.csv"])
def thermostat_ert_fu_2_ce_2(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/two_stage_ert/metadata_furnace_or_boiler_two_stage_none_single_stage.csv"])
def thermostat_ert_fu_2_na_1(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/two_stage_ert/metadata_heat_pump_electric_backup_two_stage_heat_pump_two_stage.csv"])
def thermostat_ert_hpeb_2_hp_2(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/two_stage_ert/metadata_none_two_stage_heat_pump_two_stage.csv"])
def thermostat_ert_na_2_hp_2(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session")
def core_heating_day_set_ert_hpeb_2_hp_2_entire(thermostat_ert_hpeb_2_hp_2):
    return thermostat_ert_hpeb_2_hp_2.get_core_heating_days()[0]

@pytest.fixture(scope="session")
def core_cooling_day_set_ert_hpeb_2_hp_2_entire(thermostat_ert_hpeb_2_hp_2):
    return thermostat_ert_hpeb_2_hp_2.get_core_cooling_days()[0]



VERSION = "2.0.0a6"


def get_version():
    return VERSION


# This try/except clause is a hack to make the get_version method work for the
# initial setup, which will fail with an ImportError because pandas hasn't yet
# been installed. Post-setup, this provides an import shortcut to Thermostat.
try:
    from .core import Thermostat  # noqa: F401
except ImportError:
    pass



RHU_COLUMNS = [
    'rhu1_00F_to_05F',
    'rhu1_05F_to_10F',
    'rhu1_10F_to_15F',
    'rhu1_15F_to_20F',
    'rhu1_20F_to_25F',
    'rhu1_25F_to_30F',
    'rhu1_30F_to_35F',
    'rhu1_35F_to_40F',
    'rhu1_40F_to_45F',
    'rhu1_45F_to_50F',
    'rhu1_50F_to_55F',
    'rhu1_55F_to_60F',
    'rhu1_30F_to_45F',

    'rhu2_00F_to_05F',
    'rhu2_05F_to_10F',
    'rhu2_10F_to_15F',
    'rhu2_15F_to_20F',
    'rhu2_20F_to_25F',
    'rhu2_25F_to_30F',
    'rhu2_30F_to_35F',
    'rhu2_35F_to_40F',
    'rhu2_40F_to_45F',
    'rhu2_45F_to_50F',
    'rhu2_50F_to_55F',
    'rhu2_55F_to_60F',
    'rhu2_30F_to_45F',

]

RHU2_IQFLT_COLUMNS = [
    'rhu2IQFLT_00F_to_05F',
    'rhu2IQFLT_05F_to_10F',
    'rhu2IQFLT_10F_to_15F',
    'rhu2IQFLT_15F_to_20F',
    'rhu2IQFLT_20F_to_25F',
    'rhu2IQFLT_25F_to_30F',
    'rhu2IQFLT_30F_to_35F',
    'rhu2IQFLT_35F_to_40F',
    'rhu2IQFLT_40F_to_45F',
    'rhu2IQFLT_45F_to_50F',
    'rhu2IQFLT_50F_to_55F',
    'rhu2IQFLT_55F_to_60F',
    'rhu2IQFLT_30F_to_45F',

    ]

REAL_OR_INTEGER_VALUED_COLUMNS_HEATING = [
    'n_days_in_inputfile_date_range',
    'n_days_both_heating_and_cooling',
    'n_days_insufficient_data',
    'n_core_heating_days',

    'baseline_percentile_core_heating_comfort_temperature',
    'regional_average_baseline_heating_comfort_temperature',

    'percent_savings_baseline_percentile',
    'avoided_daily_mean_core_day_runtime_baseline_percentile',
    'avoided_total_core_day_runtime_baseline_percentile',
    'baseline_daily_mean_core_day_runtime_baseline_percentile',
    'baseline_total_core_day_runtime_baseline_percentile',
    '_daily_mean_core_day_demand_baseline_baseline_percentile',
    'percent_savings_baseline_regional',
    'avoided_daily_mean_core_day_runtime_baseline_regional',
    'avoided_total_core_day_runtime_baseline_regional',
    'baseline_daily_mean_core_day_runtime_baseline_regional',
    'baseline_total_core_day_runtime_baseline_regional',
    '_daily_mean_core_day_demand_baseline_baseline_regional',
    'mean_demand',
    'alpha',
    'tau',
    'mean_sq_err',
    'root_mean_sq_err',
    'cv_root_mean_sq_err',
    'mean_abs_err',
    'mean_abs_pct_err',

    'total_core_heating_runtime',
    'total_auxiliary_heating_core_day_runtime',
    'total_emergency_heating_core_day_runtime',

    'daily_mean_core_heating_runtime',

    'core_heating_days_mean_indoor_temperature',
    'core_heating_days_mean_outdoor_temperature',
    'core_mean_indoor_temperature',
    'core_mean_outdoor_temperature',
] + RHU_COLUMNS

REAL_OR_INTEGER_VALUED_COLUMNS_COOLING = [
    'n_days_in_inputfile_date_range',
    'n_days_both_heating_and_cooling',
    'n_days_insufficient_data',
    'n_core_cooling_days',

    'baseline_percentile_core_cooling_comfort_temperature',
    'regional_average_baseline_cooling_comfort_temperature',

    'percent_savings_baseline_percentile',
    'avoided_daily_mean_core_day_runtime_baseline_percentile',
    'avoided_total_core_day_runtime_baseline_percentile',
    'baseline_daily_mean_core_day_runtime_baseline_percentile',
    'baseline_total_core_day_runtime_baseline_percentile',
    '_daily_mean_core_day_demand_baseline_baseline_percentile',
    'percent_savings_baseline_regional',
    'avoided_daily_mean_core_day_runtime_baseline_regional',
    'avoided_total_core_day_runtime_baseline_regional',
    'baseline_daily_mean_core_day_runtime_baseline_regional',
    'baseline_total_core_day_runtime_baseline_regional',
    '_daily_mean_core_day_demand_baseline_baseline_regional',
    'mean_demand',
    'alpha',
    'tau',
    'mean_sq_err',
    'root_mean_sq_err',
    'cv_root_mean_sq_err',
    'mean_abs_err',
    'mean_abs_pct_err',

    'total_core_cooling_runtime',

    'daily_mean_core_cooling_runtime',

    'core_cooling_days_mean_indoor_temperature',
    'core_cooling_days_mean_outdoor_temperature',
    'core_mean_indoor_temperature',
    'core_mean_outdoor_temperature',
]

REAL_OR_INTEGER_VALUED_COLUMNS_ALL = [
    'n_days_in_inputfile_date_range',
    'n_days_both_heating_and_cooling',
    'n_days_insufficient_data',
    'n_core_cooling_days',
    'n_core_heating_days',

    'baseline_percentile_core_cooling_comfort_temperature',
    'baseline_percentile_core_heating_comfort_temperature',
    'regional_average_baseline_cooling_comfort_temperature',
    'regional_average_baseline_heating_comfort_temperature',

    'percent_savings_baseline_percentile',
    'avoided_daily_mean_core_day_runtime_baseline_percentile',
    'avoided_total_core_day_runtime_baseline_percentile',
    'baseline_daily_mean_core_day_runtime_baseline_percentile',
    'baseline_total_core_day_runtime_baseline_percentile',
    '_daily_mean_core_day_demand_baseline_baseline_percentile',
    'percent_savings_baseline_regional',
    'avoided_daily_mean_core_day_runtime_baseline_regional',
    'avoided_total_core_day_runtime_baseline_regional',
    'baseline_daily_mean_core_day_runtime_baseline_regional',
    'baseline_total_core_day_runtime_baseline_regional',
    '_daily_mean_core_day_demand_baseline_baseline_regional',
    'mean_demand',
    'alpha',
    'tau',
    'mean_sq_err',
    'root_mean_sq_err',
    'cv_root_mean_sq_err',
    'mean_abs_err',
    'mean_abs_pct_err',

    'total_core_cooling_runtime',
    'total_core_heating_runtime',
    'total_auxiliary_heating_core_day_runtime',
    'total_emergency_heating_core_day_runtime',

    'daily_mean_core_cooling_runtime',
    'daily_mean_core_heating_runtime',

    'core_mean_indoor_temperature',
    'core_mean_outdoor_temperature',
] + RHU_COLUMNS + RHU2_IQFLT_COLUMNS

EXPORT_COLUMNS = [
    'sw_version',

    'ct_identifier',
    "heat_type",
    "heat_stage",
    "cool_type",
    "cool_stage",
    'heating_or_cooling',
    'station',
    'climate_zone',

    'start_date',
    'end_date',
    'n_days_in_inputfile_date_range',
    'n_days_both_heating_and_cooling',
    'n_days_insufficient_data',
    'n_core_cooling_days',
    'n_core_heating_days',

    'baseline_percentile_core_cooling_comfort_temperature',
    'baseline_percentile_core_heating_comfort_temperature',
    'regional_average_baseline_cooling_comfort_temperature',
    'regional_average_baseline_heating_comfort_temperature',

    'percent_savings_baseline_percentile',
    'avoided_daily_mean_core_day_runtime_baseline_percentile',
    'avoided_total_core_day_runtime_baseline_percentile',
    'baseline_daily_mean_core_day_runtime_baseline_percentile',
    'baseline_total_core_day_runtime_baseline_percentile',
    '_daily_mean_core_day_demand_baseline_baseline_percentile',
    'percent_savings_baseline_regional',
    'avoided_daily_mean_core_day_runtime_baseline_regional',
    'avoided_total_core_day_runtime_baseline_regional',
    'baseline_daily_mean_core_day_runtime_baseline_regional',
    'baseline_total_core_day_runtime_baseline_regional',
    '_daily_mean_core_day_demand_baseline_baseline_regional',
    'mean_demand',
    'alpha',
    'tau',
    'mean_sq_err',
    'root_mean_sq_err',
    'cv_root_mean_sq_err',
    'mean_abs_err',
    'mean_abs_pct_err',

    'total_core_cooling_runtime',
    'total_core_heating_runtime',
    'total_auxiliary_heating_core_day_runtime',
    'total_emergency_heating_core_day_runtime',

    'daily_mean_core_cooling_runtime',
    'daily_mean_core_heating_runtime',

    'core_cooling_days_mean_indoor_temperature',
    'core_cooling_days_mean_outdoor_temperature',
    'core_heating_days_mean_indoor_temperature',
    'core_heating_days_mean_outdoor_temperature',
    'core_mean_indoor_temperature',
    'core_mean_outdoor_temperature',
    ] + RHU_COLUMNS

CERTIFICATION_HEADERS = [
    'product_id',
    'sw_version',
    'metric',
    'filter',
    'region',
    'statistic',
    'season',
    'value',
    ]



from thermostat.importers import from_csv
from thermostat.importers import get_single_thermostat
from thermostat.util.testing import get_data_path
from thermostat.core import Thermostat, CoreDaySet
from tempfile import TemporaryDirectory

import pandas as pd
import numpy as np
from numpy import nan

import pytest

# Note:
# The following fixtures can be quite slow without a prebuilt weather cache
# they the from_csv command fetches weather data. (This happens with builds on
# travis.)
# To speed this up, spoof the weather source.

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_1_single_utc_offset_0.csv"])
def thermostat_type_1_utc(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_1_single_utc_offset_bad.csv"])
def thermostat_type_1_utc_bad(request):
    thermostats, _ = from_csv(get_data_path(request.param))

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_1_single_bad_zip.csv"])
def thermostat_type_1_zip_bad(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return list(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_multiple_same_key.csv"])
def thermostats_multiple_same_key(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return thermostats

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_1_single_too_many_minutes.csv"])
def thermostat_type_1_too_many_minutes(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return list(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_1_single_data_out_of_order.csv"])
def thermostat_type_1_data_out_of_order(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return list(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_1_single_data_missing_header.csv"])
def thermostat_type_1_data_missing_header(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return list(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_bad_equipment_types.csv"])
def thermostat_bad_equipment_type(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return list(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_1_single_metadata_missing_header.csv"])
def thermostat_type_1_metadata_missing_header(request):
    with pytest.raises(ValueError) as excinfo: 
        thermostats, _ = from_csv(get_data_path(request.param))
        return list(thermostats)
    assert "thermostat_id" in str(excinfo)
    return []

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_1_single_missing_zip.csv"])
def thermostat_type_1_zip_code_not_found(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return list(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_1_single_missing_station.csv"])
def thermostat_type_1_station_not_found(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return list(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_1_single_missing_climate_zone.csv"])
def thermostat_type_1_climate_zone_not_found(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return list(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_1_single.csv"])
def thermostat_type_1(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_1_single.csv"])
def thermostat_type_1_cache(request):
    with TemporaryDirectory() as tempdir:
        thermostats, _ = from_csv(get_data_path(request.param), save_cache=True, cache_path=tempdir)
        return next(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_2_single.csv"])
def thermostat_type_2(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_3_single.csv"])
def thermostat_type_3(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_4_single.csv"])
def thermostat_type_4(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_5_single.csv"])
def thermostat_type_5(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_single_zero_days.csv"])
def thermostat_zero_days(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_1_single_missing_over_18_days_temperature.csv"])
def thermostat_missing_over_18_days_temperature(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_1_single_missing_temperature.csv"])
def thermostat_missing_temperature(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_1_single_missing_hours.csv"])
def thermostat_missing_hours(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return next(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_type_1_single_missing_days.csv"])
def thermostat_missing_days(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return list(thermostats)

@pytest.fixture(scope="session", params=["../data/single_stage/metadata_single_emg_aux_constant_on_outlier.csv"])
def thermostat_emg_aux_constant_on_outlier(request):
    thermostats, _ = from_csv(get_data_path(request.param))
    return thermostats

@pytest.fixture(scope="session")
def core_heating_day_set_type_1_entire(thermostat_type_1):
    return thermostat_type_1.get_core_heating_days()[0]

@pytest.fixture(scope="session")
def core_heating_day_set_type_2(thermostat_type_2):
    return thermostat_type_2.get_core_heating_days()[0]

@pytest.fixture(scope="session")
def core_heating_day_set_type_3(thermostat_type_3):
    return thermostat_type_3.get_core_heating_days()[0]

@pytest.fixture(scope="session")
def core_heating_day_set_type_4(thermostat_type_4):
    return thermostat_type_4.get_core_heating_days()[0]

@pytest.fixture(scope="session")
def core_cooling_day_set_type_1_entire(thermostat_type_1):
    return thermostat_type_1.get_core_cooling_days()[0]

@pytest.fixture(scope="session")
def core_cooling_day_set_type_1_empty(thermostat_type_1):
    core_cooling_day_set = thermostat_type_1.get_core_cooling_days()[0]
    core_day_set = CoreDaySet(
        "empty",
        pd.Series(np.tile(False, core_cooling_day_set.daily.shape),
                  index=core_cooling_day_set.daily.index),
        pd.Series(np.tile(False, core_cooling_day_set.hourly.shape),
                  index=core_cooling_day_set.hourly.index),
        core_cooling_day_set.start_date,
        core_cooling_day_set.end_date
    )
    return core_day_set

@pytest.fixture(scope="session")
def core_heating_day_set_type_1_empty(thermostat_type_1):
    core_heating_day_set = thermostat_type_1.get_core_heating_days()[0]
    core_day_set = CoreDaySet(
        "empty",
        pd.Series(np.tile(False, core_heating_day_set.daily.shape),
                  index=core_heating_day_set.daily.index),
        pd.Series(np.tile(False, core_heating_day_set.hourly.shape),
                  index=core_heating_day_set.hourly.index),
        core_heating_day_set.start_date,
        core_heating_day_set.end_date
    )
    return core_day_set

@pytest.fixture(scope="session")
def core_cooling_day_set_type_2(thermostat_type_2):
    return thermostat_type_2.get_core_cooling_days()[0]

@pytest.fixture(scope="session")
def core_cooling_day_set_type_3(thermostat_type_3):
    return thermostat_type_3.get_core_cooling_days()[0]

@pytest.fixture(scope="session")
def core_cooling_day_set_type_5(thermostat_type_5):
    return thermostat_type_5.get_core_cooling_days()[0]




from thermostat.core import Thermostat
from thermostat.equipment_type import (
    has_heating,
    has_cooling,
    has_auxiliary,
    has_emergency,
    has_two_stage_cooling,
    has_two_stage_heating,
    has_multi_stage_cooling,
    has_multi_stage_heating,
    first_stage_capacity_ratio,
    )

import pandas as pd
from thermostat.zipcode_lookup import ZIPCODE_LOOKUP

from thermostat.eeweather_wrapper import get_indexed_temperatures_eeweather
from eeweather.cache import KeyValueStore
from eeweather.exceptions import ISDDataNotAvailableError
import json

import warnings
import dateutil.parser
import os
import pytz
from multiprocessing import Pool, cpu_count
from functools import partial
import logging

try:
    NUMBER_OF_CORES = len(os.sched_getaffinity(0))
except AttributeError:
    NUMBER_OF_CORES = cpu_count()
MAX_FTP_CONNECTIONS = 3
AVAILABLE_PROCESSES = min(NUMBER_OF_CORES, MAX_FTP_CONNECTIONS)

ENABLE_ENOUGH_RUNTIME_CHECK = True

logger = logging.getLogger(__name__)
warnings.simplefilter('module', Warning)

INTERVAL_COLUMNS = {
    'datetime',
    'cool_runtime_stg1',
    'cool_runtime_stg2',
    'cool_runtime_equiv',
    'heat_runtime_stg1',
    'heat_runtime_stg2',
    'heat_runtime_equiv',
    'emergency_heat_runtime',
    'auxiliary_heat_runtime',
    'temp_in',
    }


METADATA_COLUMNS = {
    'thermostat_id',
    'heat_type',
    'heat_stage',
    'cool_type',
    'cool_stage',
    'zipcode',
    'utc_offset',
    'interval_data_filename'
    }


class ZIPCodeLookupError(Exception):
    pass


class StationLookupError(Exception):
    pass


class ClimateZoneLookupError(Exception):
    pass


def _prime_eeweather_cache():
    """ Primes the eemeter / eeweather caches by doing a non-existent query
    This creates the cache directories sooner than if they were created
    during normal processing (which can lead to a race condition and missing
    thermostats)
    """
    sql_json = KeyValueStore()
    if sql_json.key_exists('0') is not False:
        raise RuntimeError("eeweather cache was not properly primed. Aborting.")


def save_json_cache(index, thermostat_id, station, cache_path=None):
    """ Saves the cached results from eeweather into a JSON file.

    Parameters
    ----------
    index : pd.DatetimeIndex
        hourly index used to compute the years needed.
    thermostat_id : str
        A unique identifier for the termostat (used for the filename)
    station : str
        Station ID used to retrieve the weather data.
    cache_path : str
        Directory path to save the cached data
    """

    json_cache = {}

    sqlite_json_store = KeyValueStore()
    years = index.groupby(index.year).keys()
    for year in years:
        base_name = f"isd-hourly-{station}-{year}"
        filename = f"{base_name}.json"
        json_cache[filename] = sqlite_json_store.retrieve_json(base_name)

    if cache_path is None:
        directory = os.path.join(
            os.curdir,
            "epathermostat_weather_data")
    else:
        directory = os.path.normpath(
            cache_path)

    thermostat_filename = f"{thermostat_id}.json"
    thermostat_path = os.path.join(directory, thermostat_filename)
    try:
        os.makedirs(os.path.dirname(directory), exist_ok=True)
        with open(thermostat_path, 'w') as outfile:
            json.dump(json_cache, outfile)

    except Exception as e:
        warnings.warn(f"Unable to write JSON file: {e}")


def normalize_utc_offset(utc_offset):
    """
    Normalizes the UTC offset
    Returns the UTC offset based on the string passed in.

    Parameters
    ----------
    utc_offset : str
        String representation of the UTC offset

    Returns
    -------
    datetime timdelta offset
    """
    try:

        if isinstance(utc_offset, int):
            symbol = ""
            if utc_offset >= 0:
                symbol = "+"
            utc_offset = symbol + str(utc_offset)

        if utc_offset == '0':
            utc_offset = '+0'

        delta = dateutil.parser.parse(
            "2000-01-01T00:00:00" + str(utc_offset)).tzinfo.utcoffset(None)
        return delta

    except (ValueError, TypeError, AttributeError) as e:
        raise TypeError(f"Invalid UTC offset: can't convert {utc_offset} to a valid offset. Please prefix with + or -. ({e})")


def from_csv(metadata_filename, verbose=False, save_cache=False, shuffle=True,
             cache_path=None,):
    """
    Creates Thermostat objects from data stored in CSV files.

    Parameters
    ----------
    metadata_filename : str
        Path to a file containing the thermostat metadata.
    verbose : boolean
        Set to True to output a more detailed log of import activity.
    save_cache: boolean
        Set to True to save the cached data to a json file (based on Thermostat ID).
    shuffle: boolean
        Shuffles the thermostats to give them random ordering if desired (helps with caching).
    cache_path: str
        Directory path to save the cached data.
    log_error: boolean
        Create a log file of thermostats that weren't imported and the reason they weren't imported.
    log_error_filename: boolean
        Name of the file to use for logging the thermostats that weren't imported.

    Returns
    -------
    thermostats : iterator over thermostat.Thermostat objects
        Thermostats imported from the given CSV input files.
    """

    _prime_eeweather_cache()

    metadata = pd.read_csv(
        metadata_filename,
        dtype={
            "thermostat_id": str,
            "heat_type": str,
            "heat_stage": str,
            "cool_type": str,
            "cool_stage": str,
            "zipcode": str,
            "utc_offset": str,
            "interval_data_filename": str
        }
    )
    metadata.fillna('', inplace=True)
    metadata.columns = map(str.lower, metadata.columns)
    if not METADATA_COLUMNS.issubset(metadata.columns):
        missing_columns = list(METADATA_COLUMNS.difference(metadata.columns))
        message = "Columns missing from metadata data file: {}".format(missing_columns)
        raise ValueError(message)

    # Shuffle the results to help alleviate cache issues
    if shuffle:
        logging.info("Metadata randomized to prevent collisions in cache.")
        metadata = metadata.sample(frac=1).reset_index(drop=True)

    p = Pool(AVAILABLE_PROCESSES)
    multiprocess_func_partial = partial(
        _multiprocess_func,
        metadata_filename=metadata_filename,
        verbose=verbose,
        save_cache=save_cache,
        cache_path=cache_path)
    result_list = p.imap(multiprocess_func_partial, metadata.iterrows())
    p.close()
    p.join()

    results = []
    error_list = []

    for result in result_list:
        if result['thermostat'] is None:
            for error in result['errors']:
                logging.warning(result['thermostat_id'] + ': ' + error)
                error_dict = {}
                error_dict['thermostat_id'] = result['thermostat_id']
                error_dict['error'] = error
            error_list.append(error_dict)
        else:
            results.append(result['thermostat'])

    # Convert this to an iterator to maintain compatibility
    return iter(results), error_list


def _multiprocess_func(metadata, metadata_filename, verbose=False, save_cache=False, cache_path=None):
    """ This function is a partial function for multiproccessing and shares the same arguments as from_csv.
    It is not intended to be called directly."""
    _, row = metadata
    logger.info(f"Importing thermostat {row.thermostat_id}")
    if verbose and logger.getEffectiveLevel() > logging.INFO:
        print(f"Importing thermostat {row.thermostat_id}")

    interval_data_filename = os.path.join(os.path.dirname(metadata_filename), row.interval_data_filename)

    status_metadata = {
        'thermostat_id': row.thermostat_id,
        'errors': [],
        'thermostat': None,
    }
    errors = []
    thermostat = None

    zipcode = row.zipcode.zfill(5)  # Ensure that we have 5 characters, and if not left-pad it with zeroes.

    try:
        thermostat = get_single_thermostat(
            thermostat_id=row.thermostat_id,
            zipcode=zipcode,
            heat_type=row.heat_type,
            heat_stage=row.heat_stage,
            cool_type=row.cool_type,
            cool_stage=row.cool_stage,
            utc_offset=row.utc_offset,
            interval_data_filename=interval_data_filename,
            save_cache=save_cache,
            cache_path=cache_path,
        )
    except (ZIPCodeLookupError, StationLookupError, ClimateZoneLookupError) as e:
        # Could not locate a station for the thermostat. Warn and skip.
        errors.append(
            "A sufficient source of outdoor weather data could not"
            f"be located for ZIP code {row.zipcode}: {e}"
            )

    except ISDDataNotAvailableError as e:
        errors.append(
            f"NCDC does not have any data: {e}")

    except (ValueError, TypeError) as e:
        errors.append(
            f"{e}")

    status_metadata['errors'] = errors
    status_metadata['thermostat'] = thermostat
    return status_metadata


def get_single_thermostat(thermostat_id, zipcode,
                          heat_type, heat_stage, cool_type, cool_stage,
                          utc_offset, interval_data_filename, save_cache=False, cache_path=None):
    """ Load a single thermostat directly from an interval data file.

    Parameters
    ----------
    thermostat_id : str
        A unique identifier for the thermostat.
    zipcode : str
        The zipcode of the thermostat, e.g. `"01234"`.
    equipment_type : str
        The equipment type of the thermostat.
    utc_offset : str
        A string representing the UTC offset of the interval data, e.g. `"-0700"`.
        Could also be `"Z"` (UTC), or just `"+7"` (equivalent to `"+0700"`),
        or any other timezone format recognized by the library
        method dateutil.parser.parse.
    interval_data_filename : str
        The path to the CSV in which the interval data is stored.
    save_cache: boolean
        Set to True to save the cached data to a json file (based on Thermostat ID).
    cache_path: str
        Directory path to save the cached data

    Returns
    -------
    thermostat : thermostat.Thermostat
        The loaded thermostat object.
    """
    # load outdoor temperatures
    zipcode_details = ZIPCODE_LOOKUP.get(zipcode)
    if zipcode_details is None:
        message = f"Could not locate ZIP Code {zipcode}"
        raise ZIPCodeLookupError(message)

    station = zipcode_details.get('station')
    if station is None:
        message = f"Could not locate a valid station for outdoor temperature " \
                "data for ZIP code {zipcode}"
        raise StationLookupError(message)

    climate_zone = zipcode_details.get("climate_zone")
    if climate_zone is None:
        message = f"Could not locate valid climate zone for ZIP Code {zipcode}"
        raise ClimateZoneLookupError(message)

    df = pd.read_csv(interval_data_filename)

    df.columns = map(str.lower, df.columns)
    if not INTERVAL_COLUMNS.issubset(df.columns):
        missing_columns = list(INTERVAL_COLUMNS.difference(df.columns))
        message = "Columns missing from interval data file: {}".format(missing_columns)
        raise ValueError(message)

    # load indices
    date_time = pd.to_datetime(df["datetime"])
    df['datetime'] = date_time
    # daily_index = pd.date_range(start=date_time[0], periods=date_time.shape[0] / 24, freq="D")
    hourly_index = pd.date_range(start=date_time[0], periods=date_time.shape[0], freq="H")
    hourly_index_utc = pd.date_range(start=date_time[0], periods=date_time.shape[0], freq="H", tz=pytz.UTC)

    # raise an error if dates are not aligned
    if not all(date_time == hourly_index):
        missing_hours = set(hourly_index).difference(set(date_time))
        duplicates = list(date_time[date_time.duplicated()])
        message = ("Dates provided for thermostat_id={} may contain some "
                   "which are out of order, missing, or duplicated.".format(thermostat_id))
        if len(duplicates) > 0:
            message += " (Possible duplicated hours: {})".format(duplicates)
        elif len(missing_hours) > 0:
            message += " (Possible missing hours: {})".format(missing_hours)
        raise ValueError(message)

    # Export the data from the cache
    if save_cache:
        save_json_cache(hourly_index, thermostat_id, station, cache_path)

    # load hourly time series values
    temp_in = _create_series(df.temp_in, hourly_index)

    utc_offset = normalize_utc_offset(utc_offset)
    temp_out = get_indexed_temperatures_eeweather(station, hourly_index_utc - utc_offset)
    temp_out.index = hourly_index

    # load daily time series values
    auxiliary_heat_runtime, emergency_heat_runtime = _calculate_aux_emerg_runtime(df, thermostat_id, heat_type, heat_stage, hourly_index)
    cool_runtime = _calculate_cool_runtime(df, thermostat_id, cool_type, cool_stage, hourly_index)
    heat_runtime = _calculate_heat_runtime(df, thermostat_id, heat_type, heat_stage, hourly_index)

    # Give the thermostats the benefit of the doubt (especially if the runtime is None)
    enough_cool_runtime = True
    enough_heat_runtime = True

    # Currently checks hourly runtime, not daily
    if cool_runtime is not None:
        cool_runtime_daily = cool_runtime.interpolate(limit=2).resample('D').agg(pd.Series.sum, skipna=False)
        enough_cool_runtime = _enough_runtime(cool_runtime_daily)
    if heat_runtime is not None:
        heat_runtime_daily = heat_runtime.interpolate(limit=2).resample('D').agg(pd.Series.sum, skipna=False)
        enough_heat_runtime = _enough_runtime(heat_runtime_daily)

    if not(enough_cool_runtime and enough_heat_runtime):
        message = "Not enough runtime for thermostat "
        if not enough_heat_runtime:
            message += "(Heat runtime has over 5% missing data) "
        if not enough_cool_runtime:
            message += "(Cool runtime has over 5% missing data) "
        raise ValueError(message)

    # create thermostat instance
    thermostat = Thermostat(
        thermostat_id,
        heat_type,
        heat_stage,
        cool_type,
        cool_stage,
        zipcode,
        station,
        climate_zone,
        temp_in,
        temp_out,
        cool_runtime,
        heat_runtime,
        auxiliary_heat_runtime,
        emergency_heat_runtime
    )
    return thermostat


def _calculate_cool_runtime(df, thermostat_id, cool_type, cool_stage, hourly_index):
    if has_cooling(cool_type):
        if df.cool_runtime_stg1.gt(60).any() or \
                df.cool_runtime_stg2.gt(60).any() or \
                df.cool_runtime_equiv.gt(60).any():
            warnings.warn("For thermostat {}, cooling runtime data was larger than 60 minutes"
                          " for one or more hours, which is impossible. Please check the data file."
                          .format(thermostat_id))
            return

        if has_multi_stage_cooling(cool_stage):
            warnings.warn("Multi-stage cooling isn't supported (yet)")
            return

        if has_two_stage_cooling(cool_stage):
            # Use cool equivalent runtime if it exists, otherwise calculate it
            if df.cool_runtime_equiv.max() > 0.0:
                cool_runtime = _create_series(df.cool_runtime_equiv, hourly_index)
            else:
                cool_runtime_both_stg = (first_stage_capacity_ratio(cool_type) * (df.cool_runtime_stg1 - df.cool_runtime_stg2)) + df.cool_runtime_stg2
                cool_runtime = _create_series(cool_runtime_both_stg, hourly_index)
        else:
            cool_runtime = _create_series(df.cool_runtime_stg1, hourly_index)
    else:
        cool_runtime = None
    return cool_runtime


def _calculate_heat_runtime(df, thermostat_id, heat_type, heat_stage, hourly_index):
    if has_heating(heat_type):
        if df.heat_runtime_stg1.gt(60).any() or \
                df.heat_runtime_stg2.gt(60).any() or \
                df.heat_runtime_equiv.gt(60).any():
            warnings.warn("For thermostat {}, heating runtime data was larger than 60 minutes"
                          " for one or more hours, which is impossible. Please check the data file."
                          .format(thermostat_id))
            return

        if has_multi_stage_heating(heat_stage):
            warnings.warn("Multi-stage heating isn't supported (yet)")
            return

        if has_two_stage_heating(heat_stage):
            # Use heat equivalent runtime if it exists, otherwise calculate it
            if df.heat_runtime_equiv.max() > 0.0:
                heat_runtime = _create_series(df.heat_runtime_equiv, hourly_index)
            else:
                heat_runtime_both_stg = (first_stage_capacity_ratio(heat_type) * (df.heat_runtime_stg1 - df.heat_runtime_stg2)) + df.heat_runtime_stg2
                heat_runtime = _create_series(heat_runtime_both_stg, hourly_index)
        else:
            heat_runtime = _create_series(df.heat_runtime_stg1, hourly_index)
    else:
        heat_runtime = None
    return heat_runtime


def _calculate_aux_emerg_runtime(df, thermostat_id, heat_type, heat_stage, hourly_index):
    if has_auxiliary(heat_type) and has_emergency(heat_type):
        auxiliary_heat_runtime = _create_series(df.auxiliary_heat_runtime, hourly_index)
        emergency_heat_runtime = _create_series(df.emergency_heat_runtime, hourly_index)
        if auxiliary_heat_runtime.gt(60).any():
            warnings.warn("For thermostat {}, auxiliary runtime data was larger than 60 minutes"
                          " for one or more hours, which is impossible. Please check the data file."
                          .format(thermostat_id))
            return
        if emergency_heat_runtime.gt(60).any():
            warnings.warn("For thermostat {}, emergency runtime data was larger than 60 minutes"
                          " for one or more hours, which is impossible. Please check the data file."
                          .format(thermostat_id))
            return
    else:
        auxiliary_heat_runtime = None
        emergency_heat_runtime = None
    return auxiliary_heat_runtime, emergency_heat_runtime


def _create_series(df, index):
    series = df
    series.index = index
    return series


def _enough_runtime(series):
    if series is None:
        return False

    if ENABLE_ENOUGH_RUNTIME_CHECK is False:
        # Don't bother checking; we're good
        return True

    num_elements = len(series)
    num_valid_elements = len(series.dropna())
    return (num_valid_elements / num_elements) > 0.95



import pytest
from numpy import nan

@pytest.fixture(scope="session")
def metrics_type_1_data():
    data = \
[{'_daily_mean_core_day_demand_baseline_baseline_percentile': 9.976082159783457,
  '_daily_mean_core_day_demand_baseline_baseline_regional': 7.12880601688671,
  'alpha': 44.06852749755634,
  'avoided_daily_mean_core_day_runtime_baseline_percentile': 192.715710435758,
  'avoided_daily_mean_core_day_runtime_baseline_regional': 67.2404434393765,
  'avoided_total_core_day_runtime_baseline_percentile': 57043.85028898437,
  'avoided_total_core_day_runtime_baseline_regional': 19903.171258055445,
  'baseline_daily_mean_core_day_runtime_baseline_percentile': 439.6312509762986,
  'baseline_daily_mean_core_day_runtime_baseline_regional': 314.155983979917,
  'baseline_percentile_core_cooling_comfort_temperature': 69.5,
  'baseline_total_core_day_runtime_baseline_percentile': 130130.85028898438,
  'baseline_total_core_day_runtime_baseline_regional': 92990.17125805543,
  'climate_zone': 'Mixed-Humid',
  'cool_stage': '',
  'cool_type': 'heat_pump',
  'core_cooling_days_mean_indoor_temperature': 73.95971753003002,
  'core_cooling_days_mean_outdoor_temperature': 79.8426321875,
  'core_mean_indoor_temperature': 73.95971753003002,
  'core_mean_outdoor_temperature': 79.8426321875,
  'ct_identifier': '8465829e-df0d-449e-97bf-96317c24dec3',
  'cv_root_mean_sq_err': 0.07887381236030344,
  'daily_mean_core_cooling_runtime': 246.91554054054055,
  'end_date': '2014-12-31T00:00:00',
  'heat_stage': 'single_stage',
  'heat_type': 'heat_pump_electric_backup',
  'heating_or_cooling': 'cooling_ALL',
  'mean_abs_err': 12.444656615631587,
  'mean_abs_pct_err': 0.050400459154527476,
  'mean_demand': 5.602990491439324,
  'mean_sq_err': 379.28224705229485,
  'n_core_cooling_days': 296,
  'n_days_both_heating_and_cooling': 212,
  'n_days_in_inputfile_date_range': 1460,
  'n_days_insufficient_data': 13,
  'percent_savings_baseline_percentile': 43.83576235942965,
  'percent_savings_baseline_regional': 21.40352145693172,
  'regional_average_baseline_cooling_comfort_temperature': 73.0,
  'root_mean_sq_err': 19.47517001343749,
  'start_date': '2011-01-01T00:00:00',
  'station': '725314',
  'sw_version': '2.0.0a6',
  'tau': -0.8165326408614967,
  'total_core_cooling_runtime': 73087.0},
 {'_daily_mean_core_day_demand_baseline_baseline_percentile': 27.33218137187926,
  '_daily_mean_core_day_demand_baseline_baseline_regional': 26.799580026817768,
  'alpha': 31.694594789454573,
  'avoided_daily_mean_core_day_runtime_baseline_percentile': 93.45842226220161,
  'avoided_daily_mean_core_day_runtime_baseline_regional': 76.57783844615918,
  'avoided_total_core_day_runtime_baseline_percentile': 83364.91265788383,
  'avoided_total_core_day_runtime_baseline_regional': 68307.43189397399,
  'baseline_daily_mean_core_day_runtime_baseline_percentile': 866.2824132935917,
  'baseline_daily_mean_core_day_runtime_baseline_regional': 849.4018294775493,
  'baseline_percentile_core_heating_comfort_temperature': 69.54999999999988,
  'baseline_total_core_day_runtime_baseline_percentile': 772723.9126578838,
  'baseline_total_core_day_runtime_baseline_regional': 757666.431893974,
  'climate_zone': 'Mixed-Humid',
  'cool_stage': '',
  'cool_type': 'heat_pump',
  'core_heating_days_mean_indoor_temperature': 66.69791666666667,
  'core_heating_days_mean_outdoor_temperature': 44.69122160033633,
  'core_mean_indoor_temperature': 66.69791666666667,
  'core_mean_outdoor_temperature': 44.69122160033633,
  'ct_identifier': '8465829e-df0d-449e-97bf-96317c24dec3',
  'cv_root_mean_sq_err': 0.10753222979305962,
  'daily_mean_core_heating_runtime': 772.8239910313902,
  'end_date': '2014-12-31T00:00:00',
  'heat_stage': 'single_stage',
  'heat_type': 'heat_pump_electric_backup',
  'heating_or_cooling': 'heating_ALL',
  'mean_abs_err': 55.66503659236822,
  'mean_abs_pct_err': 0.07202809079215974,
  'mean_demand': 24.383463368603284,
  'mean_sq_err': 6906.189550425122,
  'n_core_heating_days': 892,
  'n_days_both_heating_and_cooling': 212,
  'n_days_in_inputfile_date_range': 1460,
  'n_days_insufficient_data': 13,
  'percent_savings_baseline_percentile': 10.788447373284908,
  'percent_savings_baseline_regional': 9.015501943674966,
  'regional_average_baseline_heating_comfort_temperature': 69.0,
  'rhu1_00F_to_05F': nan,
  'rhu1_05F_to_10F': 0.35810185185185184,
  'rhu1_10F_to_15F': 0.36336805555555557,
  'rhu1_15F_to_20F': 0.37900691389063484,
  'rhu1_20F_to_25F': 0.375673443706005,
  'rhu1_25F_to_30F': 0.3318400322150354,
  'rhu1_30F_to_35F': 0.28432496802364043,
  'rhu1_30F_to_45F': 0.22154695797667256,
  'rhu1_35F_to_40F': 0.1974189826660588,
  'rhu1_40F_to_45F': 0.15271363589013506,
  'rhu1_45F_to_50F': 0.09249776186213071,
  'rhu1_50F_to_55F': 0.052322643343051506,
  'rhu1_55F_to_60F': 0.028319891645631964,
  'rhu2_00F_to_05F': nan,
  'rhu2_05F_to_10F': 0.35810185185185184,
  'rhu2_10F_to_15F': 0.36336805555555557,
  'rhu2_15F_to_20F': 0.37900691389063484,
  'rhu2_20F_to_25F': 0.375673443706005,
  'rhu2_25F_to_30F': 0.3318400322150354,
  'rhu2_30F_to_35F': 0.28432496802364043,
  'rhu2_30F_to_45F': 0.22154695797667256,
  'rhu2_35F_to_40F': 0.1974189826660588,
  'rhu2_40F_to_45F': 0.15271363589013506,
  'rhu2_45F_to_50F': 0.09249776186213071,
  'rhu2_50F_to_55F': 0.052322643343051506,
  'rhu2_55F_to_60F': 0.028319891645631964,
  'root_mean_sq_err': 83.1034869931769,
  'start_date': '2011-01-01T00:00:00',
  'station': '725314',
  'sw_version': '2.0.0a6',
  'tau': -2.351080209428112,
  'total_auxiliary_heating_core_day_runtime': 144238.0,
  'total_core_heating_runtime': 689359.0,
  'total_emergency_heating_core_day_runtime': 2104.0}]

    return data


@pytest.fixture(scope="session")
def metrics_hpeb_2_hp_2_data():
    data = \
[{'_daily_mean_core_day_demand_baseline_baseline_percentile': 12.606760008550685,
  '_daily_mean_core_day_demand_baseline_baseline_regional': 10.562788562873564,
  'alpha': 28.48315881979677,
  'avoided_daily_mean_core_day_runtime_baseline_percentile': 30.53061871305234,
  'avoided_daily_mean_core_day_runtime_baseline_regional': -27.68814459729872,
  'avoided_total_core_day_runtime_baseline_percentile': 5403.919512210265,
  'avoided_total_core_day_runtime_baseline_regional': -4900.801593721873,
  'baseline_daily_mean_core_day_runtime_baseline_percentile': 359.08034752661166,
  'baseline_daily_mean_core_day_runtime_baseline_regional': 300.86158421626055,
  'baseline_percentile_core_cooling_comfort_temperature': 70.8,
  'baseline_total_core_day_runtime_baseline_percentile': 63557.221512210264,
  'baseline_total_core_day_runtime_baseline_regional': 53252.50040627812,
  'climate_zone': 'Mixed-Humid',
  'cool_stage': 'two_stage',
  'cool_type': 'heat_pump',
  'core_cooling_days_mean_indoor_temperature': 71.87485875706216,
  'core_cooling_days_mean_outdoor_temperature': 74.53503144067797,
  'core_mean_indoor_temperature': 71.87485875706216,
  'core_mean_outdoor_temperature': 74.53503144067797,
  'ct_identifier': 'c61badb0e0c0a7e06932de804af43111',
  'cv_root_mean_sq_err': 0.16357602780022015,
  'daily_mean_core_cooling_runtime': 328.5497288135593,
  'end_date': '2018-12-31T00:00:00',
  'heat_stage': 'two_stage',
  'heat_type': 'heat_pump_electric_backup',
  'heating_or_cooling': 'cooling_ALL',
  'mean_abs_err': 40.78334959625454,
  'mean_abs_pct_err': 0.12413143588195652,
  'mean_demand': 11.534876833436256,
  'mean_sq_err': 2888.294955208049,
  'n_core_cooling_days': 177,
  'n_days_both_heating_and_cooling': 19,
  'n_days_in_inputfile_date_range': 364,
  'n_days_insufficient_data': 14,
  'percent_savings_baseline_percentile': 8.50244768986966,
  'percent_savings_baseline_regional': -9.202951140946054,
  'regional_average_baseline_cooling_comfort_temperature': 73.0,
  'root_mean_sq_err': 53.742859574161564,
  'start_date': '2018-01-01T00:00:00',
  'station': '723170',
  'sw_version': '2.0.0a6',
  'tau': 8.628075151655493,
  'total_core_cooling_runtime': 58153.301999999996},
 {'_daily_mean_core_day_demand_baseline_baseline_percentile': 14.695315526664897,
  '_daily_mean_core_day_demand_baseline_baseline_regional': 13.87197893888318,
  'alpha': 24.12888815088287,
  'avoided_daily_mean_core_day_runtime_baseline_percentile': 30.20112844422813,
  'avoided_daily_mean_core_day_runtime_baseline_regional': 10.334932007113602,
  'avoided_total_core_day_runtime_baseline_percentile': 4016.750083082341,
  'avoided_total_core_day_runtime_baseline_regional': 1374.545956946109,
  'baseline_daily_mean_core_day_runtime_baseline_percentile': 354.58162468482965,
  'baseline_daily_mean_core_day_runtime_baseline_regional': 334.7154282477151,
  'baseline_percentile_core_heating_comfort_temperature': 69.9,
  'baseline_total_core_day_runtime_baseline_percentile': 47159.35608308234,
  'baseline_total_core_day_runtime_baseline_regional': 44517.15195694611,
  'climate_zone': 'Mixed-Humid',
  'cool_stage': 'two_stage',
  'cool_type': 'heat_pump',
  'core_heating_days_mean_indoor_temperature': 68.62462406015037,
  'core_heating_days_mean_outdoor_temperature': 43.35078979323308,
  'core_mean_indoor_temperature': 68.62462406015037,
  'core_mean_outdoor_temperature': 43.35078979323308,
  'ct_identifier': 'c61badb0e0c0a7e06932de804af43111',
  'cv_root_mean_sq_err': 0.2499173059502313,
  'daily_mean_core_heating_runtime': 324.3804962406015,
  'end_date': '2018-12-31T00:00:00',
  'heat_stage': 'two_stage',
  'heat_type': 'heat_pump_electric_backup',
  'heating_or_cooling': 'heating_ALL',
  'mean_abs_err': 62.13151151808365,
  'mean_abs_pct_err': 0.19153898658567647,
  'mean_demand': 13.443657006165555,
  'mean_sq_err': 6572.069220018736,
  'n_core_heating_days': 133,
  'n_days_both_heating_and_cooling': 19,
  'n_days_in_inputfile_date_range': 364,
  'n_days_insufficient_data': 14,
  'percent_savings_baseline_percentile': 8.517398066262583,
  'percent_savings_baseline_regional': 3.0876772132131776,
  'regional_average_baseline_heating_comfort_temperature': 69.0,
  'rhu1_00F_to_05F': nan,
  'rhu1_05F_to_10F': nan,
  'rhu1_10F_to_15F': nan,
  'rhu1_15F_to_20F': 0.8938117141436386,
  'rhu1_20F_to_25F': 0.5963729624072203,
  'rhu1_25F_to_30F': 0.11484128153249978,
  'rhu1_30F_to_35F': 0.06079135092511376,
  'rhu1_30F_to_45F': 0.04260111560494782,
  'rhu1_35F_to_40F': 0.037086580256967316,
  'rhu1_40F_to_45F': 0.03371910830246602,
  'rhu1_45F_to_50F': 0.009995474641406795,
  'rhu1_50F_to_55F': 0.0,
  'rhu1_55F_to_60F': 0.012756025498260699,
  'rhu2_00F_to_05F': nan,
  'rhu2_05F_to_10F': nan,
  'rhu2_10F_to_15F': nan,
  'rhu2_15F_to_20F': 0.8938117141436386,
  'rhu2_20F_to_25F': 0.5963729624072203,
  'rhu2_25F_to_30F': nan,
  'rhu2_30F_to_35F': 0.06079135092511376,
  'rhu2_30F_to_45F': 0.04260111560494782,
  'rhu2_35F_to_40F': 0.037086580256967316,
  'rhu2_40F_to_45F': 0.03371910830246602,
  'rhu2_45F_to_50F': 0.009995474641406795,
  'rhu2_50F_to_55F': 0.0,
  'rhu2_55F_to_60F': nan,
  'root_mean_sq_err': 81.06829972325025,
  'start_date': '2018-01-01T00:00:00',
  'station': '723170',
  'sw_version': '2.0.0a6',
  'tau': 12.16945589758016,
  'total_auxiliary_heating_core_day_runtime': 1921.2,
  'total_core_heating_runtime': 43142.606,
  'total_emergency_heating_core_day_runtime': 110.0}]

    return data


@pytest.fixture(scope="session")
def metrics_ert_hpeb_2_hp_2_data():
    data = \
[{'_daily_mean_core_day_demand_baseline_baseline_percentile': 12.25281033824587,
  '_daily_mean_core_day_demand_baseline_baseline_regional': 10.312943007859884,
  'alpha': 27.60856756116322,
  'avoided_daily_mean_core_day_runtime_baseline_percentile': 27.102996583034777,
  'avoided_daily_mean_core_day_runtime_baseline_regional': -26.45396166762003,
  'avoided_total_core_day_runtime_baseline_percentile': 4770.1273986141205,
  'avoided_total_core_day_runtime_baseline_regional': -4655.897253501125,
  'baseline_daily_mean_core_day_runtime_baseline_percentile': 338.28254203758024,
  'baseline_daily_mean_core_day_runtime_baseline_regional': 284.7255837869254,
  'baseline_percentile_core_cooling_comfort_temperature': 70.9,
  'baseline_total_core_day_runtime_baseline_percentile': 59537.72739861412,
  'baseline_total_core_day_runtime_baseline_regional': 50111.70274649888,
  'climate_zone': 'Mixed-Humid',
  'cool_stage': 'two_stage',
  'cool_type': 'heat_pump',
  'core_cooling_days_mean_indoor_temperature': 71.88806818181817,
  'core_cooling_days_mean_outdoor_temperature': 74.60005241477273,
  'core_mean_indoor_temperature': 71.88806818181817,
  'core_mean_outdoor_temperature': 74.60005241477273,
  'ct_identifier': 'c61badb0e0c0a7e06932de804af43111',
  'cv_root_mean_sq_err': 0.16788677782925684,
  'daily_mean_core_cooling_runtime': 311.17954545454546,
  'end_date': '2018-12-31T00:00:00',
  'heat_stage': 'two_stage',
  'heat_type': 'heat_pump_electric_backup',
  'heating_or_cooling': 'cooling_ALL',
  'mean_abs_err': 39.55504086253661,
  'mean_abs_pct_err': 0.12711324198625543,
  'mean_demand': 11.27112244288543,
  'mean_sq_err': 2729.323861698708,
  'n_core_cooling_days': 176,
  'n_days_both_heating_and_cooling': 19,
  'n_days_in_inputfile_date_range': 364,
  'n_days_insufficient_data': 14,
  'percent_savings_baseline_percentile': 8.011940675325738,
  'percent_savings_baseline_regional': -9.291037818159984,
  'regional_average_baseline_cooling_comfort_temperature': 73.0,
  'root_mean_sq_err': 52.24293121273641,
  'start_date': '2018-01-01T00:00:00',
  'station': '723170',
  'sw_version': '2.0.0a6',
  'tau': 8.310141458385782,
  'total_core_cooling_runtime': 54767.6},
 {'_daily_mean_core_day_demand_baseline_baseline_percentile': 14.0775234005962,
  '_daily_mean_core_day_demand_baseline_baseline_regional': 13.262372650813058,
  'alpha': 23.730080239253436,
  'avoided_daily_mean_core_day_runtime_baseline_percentile': 29.61940648265708,
  'avoided_daily_mean_core_day_runtime_baseline_regional': 10.275813783215527,
  'avoided_total_core_day_runtime_baseline_percentile': 3939.3810621933917,
  'avoided_total_core_day_runtime_baseline_regional': 1366.683233167665,
  'baseline_daily_mean_core_day_runtime_baseline_percentile': 334.06075986611575,
  'baseline_daily_mean_core_day_runtime_baseline_regional': 314.7171671666742,
  'baseline_percentile_core_heating_comfort_temperature': 69.9,
  'baseline_total_core_day_runtime_baseline_percentile': 44430.08106219339,
  'baseline_total_core_day_runtime_baseline_regional': 41857.38323316767,
  'climate_zone': 'Mixed-Humid',
  'cool_stage': 'two_stage',
  'cool_type': 'heat_pump',
  'core_heating_days_mean_indoor_temperature': 68.62462406015037,
  'core_heating_days_mean_outdoor_temperature': 43.35078979323308,
  'core_mean_indoor_temperature': 68.62462406015037,
  'core_mean_outdoor_temperature': 43.35078979323308,
  'ct_identifier': 'c61badb0e0c0a7e06932de804af43111',
  'cv_root_mean_sq_err': 0.258345382172041,
  'daily_mean_core_heating_runtime': 304.44135338345876,
  'end_date': '2018-12-31T00:00:00',
  'heat_stage': 'two_stage',
  'heat_type': 'heat_pump_electric_backup',
  'heating_or_cooling': 'heating_ALL',
  'mean_abs_err': 60.50847476858618,
  'mean_abs_pct_err': 0.1987524825261593,
  'mean_demand': 12.829343614264856,
  'mean_sq_err': 6185.982599217759,
  'n_core_heating_days': 133,
  'n_days_both_heating_and_cooling': 19,
  'n_days_in_inputfile_date_range': 364,
  'n_days_insufficient_data': 14,
  'percent_savings_baseline_percentile': 8.866472822048268,
  'percent_savings_baseline_regional': 3.265094775644526,
  'regional_average_baseline_heating_comfort_temperature': 69.0,
  'rhu1_00F_to_05F': nan,
  'rhu1_05F_to_10F': nan,
  'rhu1_10F_to_15F': nan,
  'rhu1_15F_to_20F': 0.8979559976359601,
  'rhu1_20F_to_25F': 0.6045587903407809,
  'rhu1_25F_to_30F': 0.12034338820193748,
  'rhu1_30F_to_35F': 0.06371785003148996,
  'rhu1_30F_to_45F': 0.04516069332278556,
  'rhu1_35F_to_40F': 0.03934310540012056,
  'rhu1_40F_to_45F': 0.03605074276325836,
  'rhu1_45F_to_50F': 0.010863180256371055,
  'rhu1_50F_to_55F': 0.0,
  'rhu1_55F_to_60F': 0.013987335790568002,
  'rhu2_00F_to_05F': nan,
  'rhu2_05F_to_10F': nan,
  'rhu2_10F_to_15F': nan,
  'rhu2_15F_to_20F': 0.8979559976359601,
  'rhu2_20F_to_25F': 0.6045587903407809,
  'rhu2_25F_to_30F': nan,
  'rhu2_30F_to_35F': 0.06371785003148996,
  'rhu2_30F_to_45F': 0.04516069332278556,
  'rhu2_35F_to_40F': 0.03934310540012056,
  'rhu2_40F_to_45F': 0.03605074276325836,
  'rhu2_45F_to_50F': 0.010863180256371055,
  'rhu2_50F_to_55F': 0.0,
  'rhu2_55F_to_60F': nan,
  'root_mean_sq_err': 78.65101778882304,
  'start_date': '2018-01-01T00:00:00',
  'station': '723170',
  'sw_version': '2.0.0a6',
  'tau': 12.84342248121461,
  'total_auxiliary_heating_core_day_runtime': 1921.2,
  'total_core_heating_runtime': 40490.70000000001,
  'total_emergency_heating_core_day_runtime': 110.0}]

    return data




import logging
import json
import zipcodes
import warnings
from pkg_resources import resource_stream
from eeweather import (
        get_isd_file_metadata,
        rank_stations,
        select_station)
from eeweather.exceptions import (
        UnrecognizedUSAFIDError)


logging.getLogger(__name__)
warnings.simplefilter('module', Warning)

MAX_METERS = 100000

METHOD = [
        ['high', 40000],
        ['medium', 40000],
        ['high', MAX_METERS],
        ['medium', MAX_METERS],
        ]


def _zip_to_lat_long(zipcode):
    """ Returns the lat / long for a zip code, or None if none is found. """
    try:
        zip_location_list = zipcodes.matching(zipcode)
        first_location = zip_location_list.pop()
        lat = float(first_location.get('lat'))
        lon = float(first_location.get('long'))
    except ValueError:
        logging.warning(f'ZIP Code {zipcode} is invalid')
        return None, None
    except IndexError:
        logging.warning(f'No locations found for ZIP Code {zipcode}')
        return None, None
    return lat, lon


def _get_closest_station_by_zipcode_ranked(zipcode):
    """ Selects the nth ranked station from a list of ranked stations

    Parameters
    ----------
    zipcode : string
        ZIP Code

    Returns
    -------
    station : string
        Station that was found
    lat : float
        latitude for the search
    lon : float
        longitude for the search
    """

    zipcode = zipcode.zfill(5)  # Ensure that we have 5 characters, and if not left-pad it with zeroes.
    lat, lon = _zip_to_lat_long(zipcode)

    station = None

    if lat and lon:
        for min_quality, max_distance_meters in METHOD:
            station_ranking = rank_stations(lat, lon, minimum_quality=min_quality, max_distance_meters=max_distance_meters)
            # Remove airport stations like A00008
            if len(station_ranking) > 0:
                station_ranking = station_ranking[station_ranking.index.str.contains('^[0-9]')]

                station, _ = select_station(station_ranking)
                if station:
                    break

    return station, lat, lon


def get_closest_station_by_zipcode(zipcode):
    """ Look up the station by ZIP Code from eeweather
    Searches for the closest station using ZIP Codes
    The algorithm is as follows:

    1. Get the location of the ZIP Code

    2. Select the station that is the closest with the highest quality using the following method:
        * high quality stations within 40,000 meters
        * medium quality stations within 40,000 meters
        * high quality stations within 100,000 meters
        * medium quality stations within 100,000 meters

    3. If no station is returned then we log a warning message. Nothing is returned.

    4. If the station is unrecognized (:code:`UnrecognizedUSAFIDError`) then we log a warning message. Nothing is returned.

    Parameters
    ----------
    zipcode : string
        ZIP code to look up

    Returns
    -------
    station : string or None
        Station that maps to the specified ZIP code
    """

    try:
        station, lat, lon = _get_closest_station_by_zipcode_ranked(zipcode)

        if station is None:
            warnings.warn(f"No station found for ZIP {zipcode}")
            return None

        isd_metadata = get_isd_file_metadata(str(station))
        if len(isd_metadata) == 0:
            logging.warning(f"ZIP code {zipcode} mapped to station {station}, but no ISD metadata was found.")
            return None

    except UnrecognizedUSAFIDError:
        logging.warning(f"Closest station {station} is not a recognized station for ZIP {zipcode}.")
        return None

    return str(station)



from scipy.stats import norm, randint
import pandas as pd
import numpy as np
import json
from datetime import datetime

import tempfile
from itertools import islice, cycle

import pytest

from thermostat.multiple import multiple_thermostat_calculate_epa_field_savings_metrics
from thermostat.exporters import certification_to_csv
from thermostat.columns import EXPORT_COLUMNS
from thermostat.stats import (
        combine_output_dataframes,
        compute_summary_statistics,
        summary_statistics_to_csv,
        )
from .fixtures.single_stage import thermostat_emg_aux_constant_on_outlier


def get_fake_output_df(n_columns):
    columns = EXPORT_COLUMNS

    string_placeholder = ["PLACEHOLDER"] * n_columns
    zero_column = [0 if randint.rvs(0, 30) > 0 else  (None if randint.rvs(0, 2) > 0 else np.inf)
            for i in randint.rvs(0, 1, size=n_columns)]
    one_column = [1 if randint.rvs(0, 30) > 0 else  (None if randint.rvs(0, 2) > 0 else np.inf)
            for i in randint.rvs(0, 1, size=n_columns)]
    float_column = [i if randint.rvs(0, 30) > 0 else (None if randint.rvs(0, 2) > 0 else np.inf)
            for i in norm.rvs(size=n_columns)]
    zipcodes = ["01234", "12345", "23456", "34567", "43210", "54321", "65432", "76543"]
    zipcode_column = [i for i in islice(cycle(zipcodes), None, n_columns)]
    core_day_set_names = ["cooling_2012", "heating_2012-2013", "cooling_2013"]
    core_day_set_name_column = [i for i in islice(cycle(core_day_set_names), None, n_columns)]

    data = {
        'sw_version': string_placeholder,

        'ct_identifier': string_placeholder,
        'heat_type': string_placeholder,
        'heat_stage': string_placeholder,
        'cool_type': string_placeholder,
        'cool_stage': string_placeholder,
        'heating_or_cooling': core_day_set_name_column,
        'station': string_placeholder,
        'zipcode': zipcode_column,
        'climate_zone': string_placeholder,

        'start_date': datetime(2011, 1, 1),
        'end_date': datetime(2012, 1, 1),
        'n_days_both_heating_and_cooling': one_column,
        'n_days_in_inputfile_date_range': one_column,
        'n_days_insufficient_data': zero_column,
        'n_core_heating_days': one_column,

        'baseline_percentile_core_cooling_comfort_temperature': float_column,
        'baseline_percentile_core_heating_comfort_temperature': float_column,
        'regional_average_baseline_cooling_comfort_temperature': float_column,
        'regional_average_baseline_heating_comfort_temperature': float_column,

        'percent_savings_baseline_percentile': float_column,
        'avoided_daily_mean_core_day_runtime_baseline_percentile': float_column,
        'avoided_total_core_day_runtime_baseline_percentile': float_column,
        'baseline_daily_mean_core_day_runtime_baseline_percentile': float_column,
        'baseline_total_core_day_runtime_baseline_percentile': float_column,
        '_daily_mean_core_day_demand_baseline_baseline_percentile': float_column,
        'percent_savings_baseline_regional': float_column,
        'avoided_daily_mean_core_day_runtime_baseline_regional': float_column,
        'avoided_total_core_day_runtime_baseline_regional': float_column,
        'baseline_daily_mean_core_day_runtime_baseline_regional': float_column,
        'baseline_total_core_day_runtime_baseline_regional': float_column,
        '_daily_mean_core_day_demand_baseline_baseline_regional': float_column,
        'mean_demand': float_column,
        'alpha': float_column,
        'tau': float_column,
        'mean_sq_err': float_column,
        'root_mean_sq_err': float_column,
        'cv_root_mean_sq_err': float_column,
        'mean_abs_err': float_column,
        'mean_abs_pct_err': float_column,

        'total_core_cooling_runtime': float_column,
        'total_core_heating_runtime': float_column,
        'total_auxiliary_heating_core_day_runtime': float_column,
        'total_emergency_heating_core_day_runtime': float_column,

        'daily_mean_core_cooling_runtime': float_column,
        'daily_mean_core_heating_runtime': float_column,

        'core_cooling_days_mean_indoor_temperature': float_column,
        'core_cooling_days_mean_outdoor_temperature': float_column,
        'core_heating_days_mean_indoor_temperature': float_column,
        'core_heating_days_mean_outdoor_temperature': float_column,
        'core_mean_indoor_temperature': float_column,
        'core_mean_outdoor_temperature': float_column,

        'rhu1_aux_duty_cycle': float_column,
        'rhu1_emg_duty_cycle': float_column,
        'rhu1_compressor_duty_cycle': float_column,

        'rhu1_00F_to_05F': float_column,
        'rhu1_05F_to_10F': float_column,
        'rhu1_10F_to_15F': float_column,
        'rhu1_15F_to_20F': float_column,
        'rhu1_20F_to_25F': float_column,
        'rhu1_25F_to_30F': float_column,
        'rhu1_30F_to_35F': float_column,
        'rhu1_35F_to_40F': float_column,
        'rhu1_40F_to_45F': float_column,
        'rhu1_45F_to_50F': float_column,
        'rhu1_50F_to_55F': float_column,
        'rhu1_55F_to_60F': float_column,

        'rhu1_00F_to_05F_aux_duty_cycle': float_column,
        'rhu1_05F_to_10F_aux_duty_cycle': float_column,
        'rhu1_10F_to_15F_aux_duty_cycle': float_column,
        'rhu1_15F_to_20F_aux_duty_cycle': float_column,
        'rhu1_20F_to_25F_aux_duty_cycle': float_column,
        'rhu1_25F_to_30F_aux_duty_cycle': float_column,
        'rhu1_30F_to_35F_aux_duty_cycle': float_column,
        'rhu1_35F_to_40F_aux_duty_cycle': float_column,
        'rhu1_40F_to_45F_aux_duty_cycle': float_column,
        'rhu1_45F_to_50F_aux_duty_cycle': float_column,
        'rhu1_50F_to_55F_aux_duty_cycle': float_column,
        'rhu1_55F_to_60F_aux_duty_cycle': float_column,

        'rhu1_00F_to_05F_emg_duty_cycle': float_column,
        'rhu1_05F_to_10F_emg_duty_cycle': float_column,
        'rhu1_10F_to_15F_emg_duty_cycle': float_column,
        'rhu1_15F_to_20F_emg_duty_cycle': float_column,
        'rhu1_20F_to_25F_emg_duty_cycle': float_column,
        'rhu1_25F_to_30F_emg_duty_cycle': float_column,
        'rhu1_30F_to_35F_emg_duty_cycle': float_column,
        'rhu1_35F_to_40F_emg_duty_cycle': float_column,
        'rhu1_40F_to_45F_emg_duty_cycle': float_column,
        'rhu1_45F_to_50F_emg_duty_cycle': float_column,
        'rhu1_50F_to_55F_emg_duty_cycle': float_column,
        'rhu1_55F_to_60F_emg_duty_cycle': float_column,

        'rhu1_00F_to_05F_compressor_duty_cycle': float_column,
        'rhu1_05F_to_10F_compressor_duty_cycle': float_column,
        'rhu1_10F_to_15F_compressor_duty_cycle': float_column,
        'rhu1_15F_to_20F_compressor_duty_cycle': float_column,
        'rhu1_20F_to_25F_compressor_duty_cycle': float_column,
        'rhu1_25F_to_30F_compressor_duty_cycle': float_column,
        'rhu1_30F_to_35F_compressor_duty_cycle': float_column,
        'rhu1_35F_to_40F_compressor_duty_cycle': float_column,
        'rhu1_40F_to_45F_compressor_duty_cycle': float_column,
        'rhu1_45F_to_50F_compressor_duty_cycle': float_column,
        'rhu1_50F_to_55F_compressor_duty_cycle': float_column,
        'rhu1_55F_to_60F_compressor_duty_cycle': float_column,

        'rhu2_aux_duty_cycle': float_column,
        'rhu2_emg_duty_cycle': float_column,
        'rhu2_compressor_duty_cycle': float_column,

        'rhu2_00F_to_05F': float_column,
        'rhu2_05F_to_10F': float_column,
        'rhu2_10F_to_15F': float_column,
        'rhu2_15F_to_20F': float_column,
        'rhu2_20F_to_25F': float_column,
        'rhu2_25F_to_30F': float_column,
        'rhu2_30F_to_35F': float_column,
        'rhu2_35F_to_40F': float_column,
        'rhu2_40F_to_45F': float_column,
        'rhu2_45F_to_50F': float_column,
        'rhu2_50F_to_55F': float_column,
        'rhu2_55F_to_60F': float_column,

        'rhu2_00F_to_05F_aux_duty_cycle': float_column,
        'rhu2_05F_to_10F_aux_duty_cycle': float_column,
        'rhu2_10F_to_15F_aux_duty_cycle': float_column,
        'rhu2_15F_to_20F_aux_duty_cycle': float_column,
        'rhu2_20F_to_25F_aux_duty_cycle': float_column,
        'rhu2_25F_to_30F_aux_duty_cycle': float_column,
        'rhu2_30F_to_35F_aux_duty_cycle': float_column,
        'rhu2_35F_to_40F_aux_duty_cycle': float_column,
        'rhu2_40F_to_45F_aux_duty_cycle': float_column,
        'rhu2_45F_to_50F_aux_duty_cycle': float_column,
        'rhu2_50F_to_55F_aux_duty_cycle': float_column,
        'rhu2_55F_to_60F_aux_duty_cycle': float_column,

        'rhu2_00F_to_05F_emg_duty_cycle': float_column,
        'rhu2_05F_to_10F_emg_duty_cycle': float_column,
        'rhu2_10F_to_15F_emg_duty_cycle': float_column,
        'rhu2_15F_to_20F_emg_duty_cycle': float_column,
        'rhu2_20F_to_25F_emg_duty_cycle': float_column,
        'rhu2_25F_to_30F_emg_duty_cycle': float_column,
        'rhu2_30F_to_35F_emg_duty_cycle': float_column,
        'rhu2_35F_to_40F_emg_duty_cycle': float_column,
        'rhu2_40F_to_45F_emg_duty_cycle': float_column,
        'rhu2_45F_to_50F_emg_duty_cycle': float_column,
        'rhu2_50F_to_55F_emg_duty_cycle': float_column,
        'rhu2_55F_to_60F_emg_duty_cycle': float_column,

        'rhu2_00F_to_05F_compressor_duty_cycle': float_column,
        'rhu2_05F_to_10F_compressor_duty_cycle': float_column,
        'rhu2_10F_to_15F_compressor_duty_cycle': float_column,
        'rhu2_15F_to_20F_compressor_duty_cycle': float_column,
        'rhu2_20F_to_25F_compressor_duty_cycle': float_column,
        'rhu2_25F_to_30F_compressor_duty_cycle': float_column,
        'rhu2_30F_to_35F_compressor_duty_cycle': float_column,
        'rhu2_35F_to_40F_compressor_duty_cycle': float_column,
        'rhu2_40F_to_45F_compressor_duty_cycle': float_column,
        'rhu2_45F_to_50F_compressor_duty_cycle': float_column,
        'rhu2_50F_to_55F_compressor_duty_cycle': float_column,
        'rhu2_55F_to_60F_compressor_duty_cycle': float_column,

        'rhu2_30F_to_45F': float_column,
        'rhu2_30F_to_45F_aux_duty_cycle': float_column,
        'rhu2_30F_to_45F_emg_duty_cycle': float_column,
        'rhu2_30F_to_45F_compressor_duty_cycle': float_column,
    }
    df = pd.DataFrame(data, columns=columns)
    return df


@pytest.fixture
def dataframes():
    df1 = get_fake_output_df(10)
    df2 = get_fake_output_df(10)
    dfs = [df1, df2]
    return dfs


@pytest.fixture
def combined_dataframe():
    df = get_fake_output_df(100)
    return df


def test_combine_output_dataframes(dataframes):
    combined = combine_output_dataframes(dataframes)
    assert combined.shape == (20, 78)


def test_compute_summary_statistics_advanced(combined_dataframe):
    summary_statistics = compute_summary_statistics(combined_dataframe,
            advanced_filtering=True)
    assert [len(s) for s in summary_statistics] == [
            49, 49, 49, 49, 49, 49, 49, 49,
            2049, 901, 2049, 901, 2049, 901, 2049, 901,
            ]


def test_summary_statistics_to_csv(combined_dataframe):
    summary_statistics = compute_summary_statistics(combined_dataframe)


def test_none_summary_statistics_to_csv():
    assert(summary_statistics_to_csv(None, None, None) is None)


def test_compute_summary_statistics(combined_dataframe):
    summary_statistics = compute_summary_statistics(combined_dataframe)
    assert [len(s) for s in summary_statistics] == [
            49, 49, 49, 49,
            2049, 901, 2049, 901,
            ]

    _, fname = tempfile.mkstemp()
    product_id = "FAKE"
    stats_df = summary_statistics_to_csv(summary_statistics, fname, product_id)
    assert isinstance(stats_df, pd.DataFrame)

    stats_df_reread = pd.read_csv(fname)
    assert stats_df_reread.shape == (2133, 5)


def test_none_stats_certification_to_csv():
    assert(certification_to_csv(None, None, None) is None)


def test_certification(combined_dataframe):
    _, fname_stats = tempfile.mkstemp()
    _, fname_cert = tempfile.mkstemp()
    product_id = "FAKE"
    stats_df = compute_summary_statistics(combined_dataframe)
    certification_df = certification_to_csv(stats_df, fname_cert, product_id)
    assert certification_df.shape == (5, 8)


def test_bogus_method():
    with pytest.raises(ValueError):
        compute_summary_statistics(None, target_baseline_method='bogus_method')


def test_empty_metrics():
    df = pd.DataFrame([])
    assert(compute_summary_statistics(df) is None)


def test_iqr_filtering(thermostat_emg_aux_constant_on_outlier):

    thermostats_iqflt = list(thermostat_emg_aux_constant_on_outlier)
    # Run the metrics / statistics with the outlier thermostat in place
    iqflt_metrics = multiple_thermostat_calculate_epa_field_savings_metrics(thermostats_iqflt)
    iqflt_output_dataframe = pd.DataFrame(iqflt_metrics, columns=EXPORT_COLUMNS)
    iqflt_summary_statistics = compute_summary_statistics(iqflt_output_dataframe)

    # Remove the outlier thermostat
    thermostats_noiq = []
    for thermostat in list(thermostats_iqflt):
        if thermostat.thermostat_id != 'thermostat_single_emg_aux_constant_on_outlier':
            thermostats_noiq.append(thermostat)

    if len(thermostats_noiq) == 5:
        raise ValueError("Try again")

    # Re-run the metrics / statistics with the outlier thermostat removed
    noiq_metrics = multiple_thermostat_calculate_epa_field_savings_metrics(thermostats_noiq)
    noiq_output_dataframe = pd.DataFrame(noiq_metrics, columns=EXPORT_COLUMNS)
    noiq_summary_statistics = compute_summary_statistics(noiq_output_dataframe)

    # Verify that the IQFLT removed the outliers by comparing this with the
    # metrics with the outlier thermostat already removed.
    for column in range(0, len(iqflt_summary_statistics)):
        fields_iqflt = [x for x in iqflt_summary_statistics[column] if 'IQFLT' in x]
        for field_iqflt in fields_iqflt:
            field_noiq = field_iqflt.replace('rhu2IQFLT', 'rhu2')
            left_side = iqflt_summary_statistics[column][field_iqflt]
            right_side = noiq_summary_statistics[column][field_noiq]

            if np.isnan(left_side) or np.isnan(right_side):
                assert(np.isnan(left_side) and np.isnan(right_side))
            else:
                assert(left_side == right_side)



import pytest
import numpy as np
from numpy.testing import assert_allclose
from numpy import isnan
import pandas as pd

from datetime import datetime

from thermostat.importers import from_csv
from thermostat.util.testing import get_data_path

from .fixtures.two_stage import (
        thermostat_na_2_hp_2,
        thermostat_fu_2_ce_2,
        thermostat_furnace_or_boiler_two_stage_none_single_stage,
        thermostat_hpeb_2_hp_2,
        core_heating_day_set_hpeb_2_hp_2_entire,
        core_cooling_day_set_hpeb_2_hp_2_entire,
        )

from .fixtures.metrics_data import (
        metrics_hpeb_2_hp_2_data,
        )


def test_thermostat_hpeb_2_hp_2_get_core_heating_days(thermostat_hpeb_2_hp_2):
    core_heating_day_sets = thermostat_hpeb_2_hp_2.get_core_heating_days()
    assert len(core_heating_day_sets) == 1


def test_thermostat_hpeb_2_hp_2_get_core_cooling_days(thermostat_hpeb_2_hp_2):
    core_cooling_day_sets = thermostat_hpeb_2_hp_2.get_core_cooling_days()
    assert len(core_cooling_day_sets) == 1


def test_thermostat_fu_2_ce_2_get_core_heating_days(thermostat_fu_2_ce_2):
    core_heating_day_sets = thermostat_fu_2_ce_2.get_core_heating_days()
    assert len(core_heating_day_sets) == 1


def test_thermostat_fu_2_ce_2_get_core_cooling_days(thermostat_fu_2_ce_2):
    core_cooling_day_sets = thermostat_fu_2_ce_2.get_core_cooling_days()
    assert len(core_cooling_day_sets) == 1


def test_thermostat_na_2_hp_2(thermostat_na_2_hp_2):
    with pytest.raises(ValueError):
        core_heating_day_sets = thermostat_na_2_hp_2.get_core_heating_days()


def test_thermostat_na_2_hp_2(thermostat_na_2_hp_2):
    core_cooling_day_sets = thermostat_na_2_hp_2.get_core_cooling_days()
    assert len(core_cooling_day_sets) == 1


def test_thermostat_furnace_or_boiler_two_stage_none_single_stage(thermostat_furnace_or_boiler_two_stage_none_single_stage):
    with pytest.raises(ValueError):
        core_cooling_day_sets = thermostat_furnace_or_boiler_two_stage_none_single_stage.get_core_cooling_days()


def test_thermostat_furnace_or_boiler_two_stage_none_single_stage(thermostat_furnace_or_boiler_two_stage_none_single_stage):
    core_heating_day_sets = thermostat_furnace_or_boiler_two_stage_none_single_stage.get_core_heating_days()
    assert len(core_heating_day_sets) == 1


def test_thermostat_core_heating_day_set_attributes(core_heating_day_set_hpeb_2_hp_2_entire):

    assert isinstance(core_heating_day_set_hpeb_2_hp_2_entire.name, str)
    assert isinstance(core_heating_day_set_hpeb_2_hp_2_entire.daily, pd.Series)
    assert isinstance(core_heating_day_set_hpeb_2_hp_2_entire.hourly, pd.Series)
    assert core_heating_day_set_hpeb_2_hp_2_entire.daily.shape == (365,)
    assert core_heating_day_set_hpeb_2_hp_2_entire.hourly.shape == (8760,)
    assert (
        isinstance(core_heating_day_set_hpeb_2_hp_2_entire.start_date, datetime)
        or isinstance(core_heating_day_set_hpeb_2_hp_2_entire.start_date, np.datetime64)
    )
    assert (
        isinstance(core_heating_day_set_hpeb_2_hp_2_entire.end_date, datetime)
        or isinstance(core_heating_day_set_hpeb_2_hp_2_entire.end_date, np.datetime64)
    )


def test_thermostat_core_cooling_day_set_attributes(core_cooling_day_set_hpeb_2_hp_2_entire):

    assert isinstance(core_cooling_day_set_hpeb_2_hp_2_entire.name, str)
    assert isinstance(core_cooling_day_set_hpeb_2_hp_2_entire.daily, pd.Series)
    assert isinstance(core_cooling_day_set_hpeb_2_hp_2_entire.hourly, pd.Series)
    assert core_cooling_day_set_hpeb_2_hp_2_entire.daily.shape == (365,)
    assert core_cooling_day_set_hpeb_2_hp_2_entire.hourly.shape == (8760,)
    assert (
        isinstance(core_cooling_day_set_hpeb_2_hp_2_entire.start_date, datetime)
        or isinstance(core_cooling_day_set_hpeb_2_hp_2_entire.start_date, np.datetime64)
    )
    assert (
        isinstance(core_cooling_day_set_hpeb_2_hp_2_entire.end_date, datetime)
        or isinstance(core_cooling_day_set_hpeb_2_hp_2_entire.end_date, np.datetime64)
    )


def test_thermostat_hpeb_2_hp_2_total_heating_runtime(thermostat_hpeb_2_hp_2,
        core_heating_day_set_hpeb_2_hp_2_entire, metrics_hpeb_2_hp_2_data):

    total_runtime = thermostat_hpeb_2_hp_2.total_heating_runtime(core_heating_day_set_hpeb_2_hp_2_entire)
    assert_allclose(total_runtime, metrics_hpeb_2_hp_2_data[1]["total_core_heating_runtime"], rtol=1e-3)


def test_thermostat_hpeb_2_hp_2_total_emergency_heating_runtime(thermostat_hpeb_2_hp_2,
        core_heating_day_set_hpeb_2_hp_2_entire, metrics_hpeb_2_hp_2_data):

    total_runtime = thermostat_hpeb_2_hp_2.total_emergency_heating_runtime(core_heating_day_set_hpeb_2_hp_2_entire)
    assert_allclose(total_runtime, metrics_hpeb_2_hp_2_data[1]["total_emergency_heating_core_day_runtime"], rtol=1e-3)


def test_thermostat_hpeb_2_hp_2_total_auxiliary_heating_runtime(thermostat_hpeb_2_hp_2,
        core_heating_day_set_hpeb_2_hp_2_entire, metrics_hpeb_2_hp_2_data):

    total_runtime = thermostat_hpeb_2_hp_2.total_auxiliary_heating_runtime(core_heating_day_set_hpeb_2_hp_2_entire)
    assert_allclose(total_runtime, metrics_hpeb_2_hp_2_data[1]["total_auxiliary_heating_core_day_runtime"], rtol=1e-3)


def test_thermostat_hpeb_2_hp_2_total_cooling_runtime(thermostat_hpeb_2_hp_2,
        core_cooling_day_set_hpeb_2_hp_2_entire, metrics_hpeb_2_hp_2_data):

    total_runtime = thermostat_hpeb_2_hp_2.total_cooling_runtime(core_cooling_day_set_hpeb_2_hp_2_entire)
    assert_allclose(total_runtime, metrics_hpeb_2_hp_2_data[0]["total_core_cooling_runtime"], rtol=1e-3)


def test_thermostat_hpeb_2_hp_2_get_resistance_heat_utilization_bins_rhu1(thermostat_hpeb_2_hp_2,
        core_heating_day_set_hpeb_2_hp_2_entire, metrics_hpeb_2_hp_2_data):

    start = 0
    stop = 60
    step = 5
    temperature_bins = list(t for t in range(start, stop+step, step))
    rhu_runtime = thermostat_hpeb_2_hp_2.get_resistance_heat_utilization_runtime(
            core_heating_day_set_hpeb_2_hp_2_entire)
    rhu = thermostat_hpeb_2_hp_2.get_resistance_heat_utilization_bins(
            rhu_runtime,
            temperature_bins,
            core_heating_day_set_hpeb_2_hp_2_entire)

    assert len(rhu) == 12

    for item in rhu.itertuples():
        bin_name = thermostat_hpeb_2_hp_2._format_rhu('rhu1', item.Index.left, item.Index.right, duty_cycle=None)
        bin_value = item.rhu
        assert_allclose(bin_value, metrics_hpeb_2_hp_2_data[1][bin_name], rtol=1e-3)


def test_thermostat_hpeb_2_hp_2_get_resistance_heat_utilization_bins_rhu2(thermostat_hpeb_2_hp_2,
        core_heating_day_set_hpeb_2_hp_2_entire, metrics_hpeb_2_hp_2_data):

    start = 0
    stop = 60
    step = 5
    VAR_MIN_RHU_RUNTIME = 30 * 60  # Unit is in minutes (30 hours * 60 minutes)
    temperature_bins = list(t for t in range(start, stop+step, step))
    rhu_runtime = thermostat_hpeb_2_hp_2.get_resistance_heat_utilization_runtime(
            core_heating_day_set_hpeb_2_hp_2_entire)
    rhu = thermostat_hpeb_2_hp_2.get_resistance_heat_utilization_bins(
            rhu_runtime,
            temperature_bins,
            core_heating_day_set_hpeb_2_hp_2_entire,
            VAR_MIN_RHU_RUNTIME)

    assert len(rhu) == 12

    for item in rhu.itertuples():
        bin_name = thermostat_hpeb_2_hp_2._format_rhu('rhu2', item.Index.left, item.Index.right, duty_cycle=None)
        bin_value = item.rhu
        assert_allclose(bin_value, metrics_hpeb_2_hp_2_data[1][bin_name], rtol=1e-3)



@pytest.fixture(params=range(2))
def core_days(request, thermostat_hpeb_2_hp_2,
        core_heating_day_set_hpeb_2_hp_2_entire,
        core_cooling_day_set_hpeb_2_hp_2_entire):

    tests = [
        (
            thermostat_hpeb_2_hp_2,
            core_cooling_day_set_hpeb_2_hp_2_entire,
            0,
            "cooling"
        ),
        (
            thermostat_hpeb_2_hp_2,
            core_heating_day_set_hpeb_2_hp_2_entire,
            1,
            "heating"
        ),
    ]

    return tests[request.param]


def test_day_counts(core_days, metrics_hpeb_2_hp_2_data):
    thermostat, core_day_set, i, heating_or_cooling = core_days
    n_both, n_days_insufficient = thermostat.get_ignored_days(core_day_set)
    n_core_days = thermostat.get_core_day_set_n_days(core_day_set)
    n_days_in_inputfile_date_range = thermostat.get_inputfile_date_range(core_day_set)
    assert n_both == metrics_hpeb_2_hp_2_data[i]["n_days_both_heating_and_cooling"]
    assert n_days_insufficient == metrics_hpeb_2_hp_2_data[i]["n_days_insufficient_data"]
    assert n_core_days == metrics_hpeb_2_hp_2_data[i]["n_core_{}_days".format(heating_or_cooling)]
    assert n_days_in_inputfile_date_range == metrics_hpeb_2_hp_2_data[i]["n_days_in_inputfile_date_range"]



import pytest
from thermostat.stations import get_closest_station_by_zipcode, _zip_to_lat_long


def test_non_a_station():
    assert('A' not in get_closest_station_by_zipcode('77975'))


def test_bad_zipcode():
    assert(get_closest_station_by_zipcode('99999') is None)


def test_station_retrieval():
    assert('725370' in get_closest_station_by_zipcode('48242')) 


def test_no_station():
    assert(get_closest_station_by_zipcode('00602') is None)


def test_invalid_zcta():
    with pytest.warns(Warning):
        get_closest_station_by_zipcode('AAAAA')

def test_zip_to_lat_long():
    assert(_zip_to_lat_long('48242')) is not None, None




import pandas as pd
import zipcodes
from pkg_resources import resource_stream
from collections import namedtuple
from eeweather.geo import get_lat_long_climate_zones
import numpy as np
import logging

logger = logging.getLogger('epathermostat')

BASELINE_TEMPERATURE = {
    'Very-Cold/Cold': {
        'heating': 68.0,
        'cooling': 73.0,
        },
    'Mixed-Humid': {
        'heating': 69.0,
        'cooling': 73.0,
        },
    'Mixed-Dry/Hot-Dry': {
        'heating': 69.0,
        'cooling': 75.0,
        },
    'Hot-Humid': {
        'heating': 70.0,
        'cooling': 75.0,
        },
    'Marine': {
        'heating': 67.0,
        'cooling': np.nan,
        }
    }

CLIMATE_ZONE_MAPPING = {
    'Cold': 'Very-Cold/Cold',
    'Very Cold': 'Very-Cold/Cold',
    'Hot-Dry': 'Mixed-Dry/Hot-Dry',
    'Mixed-Dry': 'Mixed-Dry/Hot-Dry',
    }


HEATING_CLIMATE_ZONE_WEIGHTS = {
    'very-cold_cold': 0.549,
    'mixed-humid': 0.312,
    'mixed-dry_hot-dry': 0.054,
    'hot-humid': 0.049,
    'marine': 0.036}

COOLING_CLIMATE_ZONE_WEIGHTS = {
    'very-cold_cold': 0.096,
    'mixed-humid': 0.34,
    'mixed-dry_hot-dry': 0.144,
    'hot-humid': 0.42,
    'marine': 0.0}


def retrieve_climate_zone(zipcode):
    """ Performs a lookup of the Climate Zone from eeweather
    and returns the climate zone and baseline regional comfort temperatures.

    Parameters
    ----------
    zipcode : The ZIP Code to lookup using eeweather's climate zones

    Returns
    -------

    climate_zone_nt : named tuple
       Named Tuple consisting of the Climate Zone, baseline_regional_cooling_comfort_temperature, and baseline_regional_heating_comfort_temperature
    """
    ClimateZone = namedtuple('ClimateZone', ['climate_zone', 'baseline_regional_cooling_comfort_temperature', 'baseline_regional_heating_comfort_temperature'])
    try:
        zipcode = zipcode.zfill(5)
        zipcode_details = zipcodes.matching(zipcode).pop()
        latitude = float(zipcode_details['lat'])
        longitude = float(zipcode_details['long'])
        ee_climate_zones = get_lat_long_climate_zones(latitude, longitude)
        ba_climate_zone = ee_climate_zones['ba_climate_zone']
        climate_zone = CLIMATE_ZONE_MAPPING.get(ba_climate_zone, ba_climate_zone)
        baseline_regional_cooling_comfort_temperature = BASELINE_TEMPERATURE.get(climate_zone, {}).get('cooling', None)
        baseline_regional_heating_comfort_temperature = BASELINE_TEMPERATURE.get(climate_zone, {}).get('heating', None)

        climate_zone_nt = ClimateZone(climate_zone, baseline_regional_cooling_comfort_temperature, baseline_regional_heating_comfort_temperature)
    except IndexError:
        logger.warning(f'ZIP Code {zipcode} is not found. Is it valid?')
        climate_zone_nt = ClimateZone(np.nan, np.nan, np.nan)
    return climate_zone_nt



import logging

HEAT_TYPE = [
    'furnace_or_boiler',  # Non heat pump heating (gas or oil furnace, electric resistance)
    'heat_pump_electric_backup',   # Heat pump with electric resistance heat (strip heat)
    'heat_pump_no_electric_backup',  # Heat pump without electric resistance heat
    'heat_pump_dual_fuel',  # Dual fuel heat pump (e.g. gas or oil fired)
    'other',  # Multi-zone, ?
    'none',  # No central heating system
    ]


HEAT_STAGE = [
    'single_stage',  # Single capacity heater or single stage compressor
    'single_speed',  # Single capacity heater or single stage compressor (synonym)
    'two_stage',  # Dual capacity heater or dual stage compressor
    'two_speed',  # Dual capacity heater or dual stage compressor (synonym)
    'modulating',  # Modulating or variable capacity unit
    'variable_speed',   # Modulating or variable capacity unit
    ]

COOL_TYPE = [
    'heat_pump',  # Heat pump w/ cooling
    'central',  # Central AC
    'other',  # Mini-split, evaporative cooler, ?
    'none',  # No central cooling system
    ]

COOL_STAGE = [
    'single_speed',  # Single stage compressor
    'single_stage',  # Single stage compressor (synonym)
    'two_speed',  # Dual stage compressor
    'two_stage',  # Dual stage compressor (synonym)
    'modulating',  # Modulating or variable capacity compressor
    'variable_speed',   # Modulating or variable capacity unit
    ]

#: This mapping is for old scripts that need to refer to the old mapping, but want to use the new functionality.
#:
#: Usage:
#:
#: ``equipment = EQUIPMENT_MAPPING[equipment_type]``
#:
#: ``heat_type = equipment['heat_type']``
#:
#: Note: Newer scripts should explicitly set the ``heat_type``, ``cool_type``, ``heat_stage``, and ``cool_stage``
#:
#:    ``equipment_type : { 0, 1, 2, 3, 4, 5 }``
#:        - :code:`0`: Other - e.g. multi-zone multi-stage, modulating. Note: module will
#:          not output savings data for this type.
#:        - :code:`1`: Single stage heat pump with aux and/or emergency heat
#:        - :code:`2`: Single stage heat pump without aux or emergency heat
#:        - :code:`3`: Single stage non heat pump with single-stage central air conditioning
#:        - :code:`4`: Single stage non heat pump without central air conditioning
#:        - :code:`5`: Single stage central air conditioning without central heating
EQUIPMENT_MAPPING = [
        {'heat_type': None, 'heat_stage': 'two_stage', 'cool_type': None, 'cool_stage': 'two_speed'},  # 0
        {'heat_type': 'heat_pump_electric_backup', 'heat_stage': 'single_stage', 'cool_type': 'heat_pump', 'cool_stage': None},  # 1
        {'heat_type': 'heat_pump_no_electric_backup', 'heat_stage': 'single_stage', 'cool_type': 'heat_pump', 'cool_stage': None},  # 2
        {'heat_type': 'furnace_or_boiler', 'heat_stage': 'single_stage', 'cool_type': 'central', 'cool_stage': 'single_speed'},  # 3
        {'heat_type': 'furnace_or_boiler', 'heat_stage': 'single_stage', 'cool_type': 'none', 'cool_stage': None},  # 4
        {'heat_type': 'none', 'heat_stage': None, 'cool_type': 'central', 'cool_stage': 'single_speed'},  # 5
        ]


def has_heating(heat_type):
    """ Determines of the heat type has heating capability

    Parameters
    ----------
    heat_type: str
        The name of the heat type

    Returns
    -------
    boolean
    """
    if heat_type is None:
        return False
    if heat_type == 'none':
        return False
    if heat_type == 'other':
        return False
    if heat_type == 'heat_pump_dual_fuel':
        return False
    if heat_type in HEAT_TYPE:
        return True
    return False


def has_cooling(cool_type):
    """ Determines if the cooling type has cooling capability

    Parameters
    ----------
    cool_type : str
        The name of the heat type


    Returns
    -------
    boolean
    """
    if cool_type is None:
        return False
    if cool_type == 'none':
        return False
    if cool_type == 'other':
        return False
    if cool_type in COOL_TYPE:
        return True
    return False


def has_auxiliary(heat_type):
    """ Determines if the heating type has aux capability

    Parameters
    ----------
    heat_type : str
        The name of the heat type


    Returns
    -------
    boolean
    """
    if heat_type == 'heat_pump_electric_backup':
        return True
    return False


def has_emergency(heat_type):
    """ Determines if the heating type has emergency capability

    Parameters
    ----------
    heat_type : str
        The name of the heat type


    Returns
    -------
    boolean
    """
    if heat_type == 'heat_pump_electric_backup':
        return True
    return False


def has_two_stage_heating(heat_stage):
    """ Determines if the heating stage has two-stage capability

    Parameters
    ----------
    heat_stage : str
        The name of the heat stage


    Returns
    -------
    boolean
    """
    if heat_stage == 'two_stage' or heat_stage == 'two_speed':
        return True
    return False


def has_two_stage_cooling(cool_stage):
    """ Determines if the cooling stage has two-stage capability

    Parameters
    ----------
    cool_stage : str
        The name of the cooling stage


    Returns
    -------
    boolean
    """
    if cool_stage == 'two_speed' or cool_stage == 'two_stage':
        return True
    return False


def has_multi_stage_cooling(cool_stage):
    """ Determines if the cooling stage has multi-stage capability

    Parameters
    ----------
    cool_stage : str
        The name of the cooling stage


    Returns
    -------
    boolean
    """
    if cool_stage == 'variable_speed' or cool_stage == 'modulating':
        return True
    return False


def has_multi_stage_heating(heat_stage):
    """ Determines if the heating stage has multi-stage capability

    Parameters
    ----------
    cool_stage : str
        The name of the cooling stage


    Returns
    -------
    boolean
    """
    if heat_stage == 'variable_speed' or heat_stage == 'modulating':
        return True
    return False


def has_resistance_heat(heat_type):
    """ Determines if the heat type has resistance heating capability

    Parameters
    ----------
    heat_type : str
        The name of the heating type


    Returns
    -------
    boolean
    """
    if heat_type == 'heat_pump_electric_backup':
        return True
    return False


def validate_heat_type(heat_type):
    if heat_type is None or heat_type == '' or heat_type in HEAT_TYPE:
        return True
    logging.warning("Heating type {heat_type} is not a recognized heating type.".format(heat_type=heat_type))
    return False


def validate_cool_type(cool_type):
    if cool_type is None or cool_type == '' or cool_type in COOL_TYPE:
        return True
    logging.warning("Cooling type {cool_type} is not a recognized cooling type.".format(cool_type=cool_type))
    return False


def validate_heat_stage(heat_stage):
    if heat_stage is None or heat_stage == '' or heat_stage in HEAT_STAGE:
        return True
    logging.warning("Heating stage {heat_stage} is not a recognized heating stage.".format(heat_stage=heat_stage))
    return False


def validate_cool_stage(cool_stage):
    if cool_stage is None or cool_stage == '' or cool_stage in COOL_STAGE:
        return True
    logging.warning("Cooling stage {cool_stage} is not a recognized cooling stage.".format(cool_stage=cool_stage))
    return False


def first_stage_capacity_ratio(heat_or_cool_type):
    if heat_or_cool_type == "furnace_or_boiler":
        return 0.65
    else:
        return 0.72



import datetime
import pandas as pd
import pytest

from thermostat.importers import (
        from_csv,
        normalize_utc_offset,
        )

from thermostat.util.testing import get_data_path

from .fixtures.two_stage import (
        thermostat_hpeb_2_hp_2,
        )

def test_import_csv(thermostat_hpeb_2_hp_2):

    def assert_is_series_with_shape(series, shape):
        assert isinstance(series, pd.Series)
        assert series.shape == shape

    assert_is_series_with_shape(thermostat_hpeb_2_hp_2.cool_runtime_daily, (365,))
    assert_is_series_with_shape(thermostat_hpeb_2_hp_2.heat_runtime_daily, (365,))

    assert_is_series_with_shape(thermostat_hpeb_2_hp_2.cool_runtime_hourly, (8760,))
    assert_is_series_with_shape(thermostat_hpeb_2_hp_2.heat_runtime_hourly, (8760,))

    assert_is_series_with_shape(thermostat_hpeb_2_hp_2.auxiliary_heat_runtime, (8760,))
    assert_is_series_with_shape(thermostat_hpeb_2_hp_2.emergency_heat_runtime, (8760,))

    assert_is_series_with_shape(thermostat_hpeb_2_hp_2.temperature_in, (8760,))
    assert_is_series_with_shape(thermostat_hpeb_2_hp_2.temperature_out, (8760,))



import pandas as pd
import numpy as np
import sys
import os
import uuid

from collections import OrderedDict
from itertools import chain
import warnings
from functools import reduce
from pkg_resources import resource_stream
import logging

from functools import partial
from itertools import repeat
import multiprocessing

if os.name == 'nt':
    from multiprocessing.pool import ThreadPool as Pool
else:
    from multiprocessing import Pool

from thermostat import get_version
from thermostat.columns import (
        REAL_OR_INTEGER_VALUED_COLUMNS_HEATING,
        REAL_OR_INTEGER_VALUED_COLUMNS_COOLING,
        REAL_OR_INTEGER_VALUED_COLUMNS_ALL
        )
from thermostat.climate_zone import HEATING_CLIMATE_ZONE_WEIGHTS, COOLING_CLIMATE_ZONE_WEIGHTS

QUANTILE = [1, 2.5, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95, 98, 99]
IQR_FILTER_PARAMETER = 1.5
TOP_ONLY_PERCENTILE_FILTER = .05  # Filters top 5 percent for RHU2 calculation
UNFILTERED_PERCENTILE = 1 - TOP_ONLY_PERCENTILE_FILTER

logger = logging.getLogger('epathermostat')
warnings.simplefilter('module', Warning)

target_baseline_method = 'baseline_percentile'

def globalize(func):
    def result(*args, **kwargs):
        return func(*args, **kwargs)
    result.__name__ = result.__qualname__ = uuid.uuid4().hex
    setattr(sys.modules[result.__module__], result.__name__, result)
    return result


def combine_output_dataframes(dfs):
    """ Combines output dataframes. Useful when combining output from batches.

    Parameters
    ----------
    dfs : list of pd.DataFrame
        Output dataFrames to combine into one.

    Returns
    -------
    out : pd.DataFrame
        Dataframe with combined output metadata.

    """
    return pd.concat(dfs, ignore_index=True)


def get_filtered_stats(
        df, row_filter, label, heating_or_cooling, target_columns,
        target_baseline_method):

    n_rows_total = df.shape[0]

    filtered_df = df[[row_filter(row, df) for i, row in df.iterrows()]]

    n_rows_kept = filtered_df.shape[0]
    n_rows_discarded = n_rows_total - n_rows_kept

    stats = OrderedDict()
    stats["label"] = "{}_{}".format(label, heating_or_cooling)
    stats["sw_version"] = get_version()
    stats["n_thermostat_core_day_sets_total"] = n_rows_total
    stats["n_thermostat_core_day_sets_kept"] = n_rows_kept
    stats["n_thermostat_core_day_sets_discarded"] = n_rows_discarded

    if n_rows_total == 0:
        warnings.warn(
            "Not enough data to compute summary_statistics ({}_{})"
            .format(label, heating_or_cooling)
        )
        return []

    for column_name in target_columns:
        column = filtered_df[column_name].replace([np.inf, -np.inf], np.nan).dropna()

        # calculate quantiles and statistics
        col_num = pd.to_numeric(column)
        if col_num.empty:
            mean = np.nan
            sem = np.nan
            lower_bound = np.nan
            upper_bound = np.nan
        else:
            mean = np.nanmean(col_num)
            sem = np.nanstd(column) / (column.count() ** .5)
            lower_bound = mean - (1.96 * sem)
            upper_bound = mean + (1.96 * sem)

        stats["{}_n".format(column_name)] = column.count()
        stats["{}_upper_bound_95_perc_conf".format(column_name)] = upper_bound
        stats["{}_mean".format(column_name)] = mean
        stats["{}_lower_bound_95_perc_conf".format(column_name)] = lower_bound
        stats["{}_sem".format(column_name)] = sem

        for quantile in QUANTILE:
            stats["{}_q{}".format(column_name, quantile)] = column.quantile(quantile / 100.)

        # Calculate IQR for RHU2 and filter outliers
        # This adds an rhu2IQFLT version of the rhu2 column to the output data.
        if 'rhu2_' in column_name:
            iqr_column_name = column_name.replace('rhu2', 'rhu2IQFLT')
            iqr_filter = (column < column.quantile(UNFILTERED_PERCENTILE))
            if bool(iqr_filter.any()) is False:
                iqr_filter = (column == column)
                logging.debug("RHU filtering 5% and min Runtime filtering removed entire dataset from statistics summary for bin. Disabling filter.")
            iqr_filtered_column = column.loc[iqr_filter]

            # calculate quantiles and statistics for RHU2 IQR (IQFLT) and
            # non-IQR filtering (NOIQ)
            iqr_num = pd.to_numeric(iqr_filtered_column)
            if iqr_num.empty:
                iqr_mean = np.nan
                iqr_sem = np.nan
                iqr_lower_bound = np.nan
                iqr_upper_bound = np.nan
            else:
                iqr_mean = np.nanmean(iqr_num)
                iqr_sem = np.nanstd(iqr_filtered_column) / (iqr_filtered_column.count() ** .5)
                iqr_lower_bound = iqr_mean - (1.96 * iqr_sem)
                iqr_upper_bound = iqr_mean + (1.96 * iqr_sem)

            stats["{}_n".format(iqr_column_name)] = iqr_filtered_column.count()
            stats["{}_upper_bound_95_perc_conf".format(iqr_column_name)] = iqr_upper_bound
            stats["{}_mean".format(iqr_column_name)] = iqr_mean
            stats["{}_lower_bound_95_perc_conf".format(iqr_column_name)] = iqr_lower_bound
            stats["{}_sem".format(iqr_column_name)] = iqr_sem

            for quantile in QUANTILE:
                stats["{}_q{}".format(iqr_column_name, quantile)] = iqr_filtered_column.quantile(quantile / 100.)

    return stats


def heating_stats(df, filter_, label, target_baseline_method):
    heating_df = df[["heating" in name for name in df["heating_or_cooling"]]]
    return get_filtered_stats(
        heating_df, filter_, label,
        "heating", REAL_OR_INTEGER_VALUED_COLUMNS_HEATING,
        target_baseline_method)


def cooling_stats(df, filter_, label, target_baseline_method):
    cooling_df = df[["cooling" in name for name in df["heating_or_cooling"]]]
    return get_filtered_stats(
        cooling_df, filter_, label,
        "cooling", REAL_OR_INTEGER_VALUED_COLUMNS_COOLING,
        target_baseline_method)


def _identity_filter(row, df):
    return True


def _range_filter(row, column_name, heating_or_cooling, lower_bound=-np.inf, upper_bound=np.inf, target_baseline=False):
    if target_baseline:
        full_column_selector = "{}_{}".format(column_name, target_baseline_method)
    else:
        full_column_selector = column_name
    column_value = row[full_column_selector]
    return lower_bound < column_value < upper_bound

def _percentile_range_filter(row, column_name, heating_or_cooling, df, quantile=0.0, target_baseline=False):
    if target_baseline:
        full_column_selector = "{}_{}".format(column_name, target_baseline_method)
    else:
        full_column_selector = column_name
    lower_bound = df[full_column_selector].dropna().quantile(0.0 + quantile)
    upper_bound = df[full_column_selector].dropna().quantile(1.0 - quantile)
    return _range_filter(row, column_name, heating_or_cooling, lower_bound, upper_bound, target_baseline)


def _tau_filter_heating(row, df):
    return _range_filter(row, "tau", "heating", 0, 25)


def _tau_filter_cooling(row, df):
    return _range_filter(row, "tau", "cooling", 0, 25)


def _cvrmse_filter_heating(row, df):
    return _range_filter(row, "cv_root_mean_sq_err", "heating", upper_bound=0.6)


def _cvrmse_filter_cooling(row, df):
    return _range_filter(row, "cv_root_mean_sq_err", "cooling", upper_bound=0.6)


def _savings_filter_p01_heating(row, df):
    return _percentile_range_filter(row, "percent_savings", "heating", df, 0.01, True)


def _savings_filter_p01_cooling(row, df):
    return _percentile_range_filter(row, "percent_savings", "cooling", df, 0.01, True)


def _combine_filters(filters):
    @globalize
    def _new_filter(row, df):
        return reduce(lambda x, y: x and y(row, df), filters, True)
    return _new_filter


def compute_summary_statistics(
        metrics_df,
        target_baseline_method="baseline_percentile",
        advanced_filtering=False):
    """ Computes summary statistics for the output dataframe. Computes the
    following statistics for each real-valued or integer valued column in
    the output dataframe: mean, standard error of the mean, and deciles.

    Parameters
    ----------
    df : pd.DataFrame
        Output for which to compute summary statistics.
    label : str
        Name for this set of thermostat outputs.
    target_baseline_method : {"baseline_percentile", "baseline_regional"}, default "baseline_percentile"
        Baselining method by which samples will be filtered according to bad fits.

    Returns
    -------
    stats : collections.OrderedDict
        An ordered dict containing the summary statistics. Column names are as
        follows, in which ### is a placeholder for the name of the column:

          - mean: ###_mean
          - standard error of the mean: ###_sem
          - 1st quantile:  ###_1q
          - 2.5th quantile:###_2.5q
          - 5th quantile:  ###_5q
          - 10th quantile: ###_10q
          - 15th quantile: ###_15q
          - 20th quantile: ###_20q
          - 25th quantile: ###_25q
          - 30th quantile: ###_30q
          - 35th quantile: ###_35q
          - 40th quantile: ###_40q
          - 45th quantile: ###_45q
          - 50th quantile: ###_50q
          - 55th quantile: ###_55q
          - 60th quantile: ###_60q
          - 65th quantile: ###_65q
          - 70th quantile: ###_70q
          - 75th quantile: ###_75q
          - 80th quantile: ###_80q
          - 85th quantile: ###_85q
          - 90th quantile: ###_90q
          - 95th quantile: ###_95q
          - 98th quantile: ###_98q
          - 99th quantile: ###_99q
          - number of non-null core day sets: ###_n

        The following general values are also output:

          - label: label
          - number of total core day sets: n_total_core_day_sets

    """

    if target_baseline_method not in ["baseline_percentile", "baseline_regional"]:
        message = (
            'Baseline method not supported - please use one of'
            ' "baseline_percentile" or "baseline_regional"'
        )
        raise ValueError(message)

    if len(metrics_df) == 0:
        warnings.warn("No data to compute for summary statistics.")
        return None

    very_cold_cold_df = metrics_df[[
        (cz is not None) and "Very-Cold/Cold" in cz
        for cz in metrics_df["climate_zone"]
    ]]
    mixed_humid_df = metrics_df[[
        (cz is not None) and "Mixed-Humid" in cz
        for cz in metrics_df["climate_zone"]
    ]]
    mixed_dry_hot_dry_df = metrics_df[[
        (cz is not None) and "Mixed-Dry/Hot-Dry" in cz
        for cz in metrics_df["climate_zone"]
    ]]
    hot_humid_df = metrics_df[[
        (cz is not None) and "Hot-Humid" in cz
        for cz in metrics_df["climate_zone"]
    ]]
    marine_df = metrics_df[[
        (cz is not None) and "Marine" in cz
        for cz in metrics_df["climate_zone"]
    ]]

    filter_0 = _identity_filter
    filter_1_heating = _combine_filters([_tau_filter_heating])
    filter_1_cooling = _combine_filters([_tau_filter_cooling])
    filter_2_heating = _combine_filters([_tau_filter_heating, _cvrmse_filter_heating])
    filter_2_cooling = _combine_filters([_tau_filter_cooling, _cvrmse_filter_cooling])
    filter_3_heating = _combine_filters([_tau_filter_heating, _cvrmse_filter_heating, _savings_filter_p01_heating])
    filter_3_cooling = _combine_filters([_tau_filter_cooling, _cvrmse_filter_cooling, _savings_filter_p01_cooling])

    no_filter_list = [
            (metrics_df, filter_0, "all_no_filter", target_baseline_method),
            (very_cold_cold_df, filter_0, "very-cold_cold_no_filter", target_baseline_method),
            (mixed_humid_df, filter_0, "mixed-humid_no_filter", target_baseline_method),
            (mixed_dry_hot_dry_df, filter_0, "mixed-dry_hot-dry_no_filter", target_baseline_method),
            (hot_humid_df, filter_0, "hot-humid_no_filter", target_baseline_method),
            (marine_df, filter_0, "marine_no_filter", target_baseline_method),
            ]

    filter_1_heating_list = [
            (metrics_df, filter_1_heating, "all_tau_filter", target_baseline_method),
            (very_cold_cold_df, filter_1_heating, "very-cold_cold_tau_filter", target_baseline_method),
            (mixed_humid_df, filter_1_heating, "mixed-humid_tau_filter", target_baseline_method),
            (mixed_dry_hot_dry_df, filter_1_heating, "mixed-dry_hot-dry_tau_filter", target_baseline_method),
            (hot_humid_df, filter_1_heating, "hot-humid_tau_filter", target_baseline_method),
            (marine_df, filter_1_heating, "marine_tau_filter", target_baseline_method),
            ]

    filter_1_cooling_list = [
            (metrics_df, filter_1_cooling, "all_tau_filter", target_baseline_method),
            (very_cold_cold_df, filter_1_cooling, "very-cold_cold_tau_filter", target_baseline_method),
            (mixed_humid_df, filter_1_cooling, "mixed-humid_tau_filter", target_baseline_method),
            (mixed_dry_hot_dry_df, filter_1_cooling, "mixed-dry_hot-dry_tau_filter", target_baseline_method),
            (hot_humid_df, filter_1_cooling, "hot-humid_tau_filter", target_baseline_method),
            (marine_df, filter_1_cooling, "marine_tau_filter", target_baseline_method),
            ]

    filter_2_heating_list = [
            (metrics_df, filter_2_heating, "all_cvrmse_filter", target_baseline_method),
            (very_cold_cold_df, filter_2_heating, "very-cold_cold_cvrmse_filter", target_baseline_method),
            (mixed_humid_df, filter_2_heating, "mixed-humid_cvrmse_filter", target_baseline_method),
            (mixed_dry_hot_dry_df, filter_2_heating, "mixed-dry_hot-dry_cvrmse_filter", target_baseline_method),
            (hot_humid_df, filter_2_heating, "hot-humid_cvrmse_filter", target_baseline_method),
            (marine_df, filter_2_heating, "marine_cvrmse_filter", target_baseline_method),
            ]

    filter_2_cooling_list = [
            (metrics_df, filter_2_cooling, "all_cvrmse_filter", target_baseline_method),
            (very_cold_cold_df, filter_2_cooling, "very-cold_cold_cvrmse_filter", target_baseline_method),
            (mixed_humid_df, filter_2_cooling, "mixed-humid_cvrmse_filter", target_baseline_method),
            (mixed_dry_hot_dry_df, filter_2_cooling, "mixed-dry_hot-dry_cvrmse_filter", target_baseline_method),
            (hot_humid_df, filter_2_cooling, "hot-humid_cvrmse_filter", target_baseline_method),
            (marine_df, filter_2_cooling, "marine_cvrmse_filter", target_baseline_method),
            ]

    filter_3_heating_list = [
            (metrics_df, filter_3_heating, "all_tau_cvrmse_savings_p01_filter", target_baseline_method),
            (very_cold_cold_df, filter_3_heating, "very-cold_cold_tau_cvrmse_savings_p01_filter", target_baseline_method),
            (mixed_humid_df, filter_3_heating, "mixed-humid_tau_cvrmse_savings_p01_filter", target_baseline_method),
            (mixed_dry_hot_dry_df, filter_3_heating, "mixed-dry_hot-dry_tau_cvrmse_savings_p01_filter", target_baseline_method),
            (hot_humid_df, filter_3_heating, "hot-humid_tau_cvrmse_savings_p01_filter", target_baseline_method),
            (marine_df, filter_3_heating, "marine_tau_cvrmse_savings_p01_filter", target_baseline_method),
            ]

    filter_3_cooling_list = [
            (metrics_df, filter_3_cooling, "all_tau_cvrmse_savings_p01_filter", target_baseline_method),
            (very_cold_cold_df, filter_3_cooling, "very-cold_cold_tau_cvrmse_savings_p01_filter", target_baseline_method),
            (mixed_humid_df, filter_3_cooling, "mixed-humid_tau_cvrmse_savings_p01_filter", target_baseline_method),
            (mixed_dry_hot_dry_df, filter_3_cooling, "mixed-dry_hot-dry_tau_cvrmse_savings_p01_filter", target_baseline_method),
            (hot_humid_df, filter_3_cooling, "hot-humid_tau_cvrmse_savings_p01_filter", target_baseline_method),
            (marine_df, filter_3_cooling, "marine_tau_cvrmse_savings_p01_filter", target_baseline_method),
            ]

    # Windows needs this
    stats_dict = {}
    with Pool() as pool:
        heating_stats_list = pool.starmap(heating_stats, no_filter_list)
        cooling_stats_list = pool.starmap(cooling_stats, no_filter_list)
        heating_stats_filter3_list = pool.starmap(heating_stats, filter_3_heating_list)
        cooling_stats_filter3_list = pool.starmap(cooling_stats, filter_3_cooling_list)

    # Interleave the lists (all_heating, all_cooling, very-cold_heating...)
    no_filter_list = list(chain(*zip(heating_stats_list, cooling_stats_list)))
    filter_3_list = list(chain(*zip(heating_stats_filter3_list, cooling_stats_filter3_list)))

    if advanced_filtering:
        with Pool() as pool:
            heating_stats_filter1_list = pool.starmap(heating_stats, filter_1_heating_list)
            cooling_stats_filter1_list = pool.starmap(cooling_stats, filter_1_cooling_list)
            heating_stats_filter2_list = pool.starmap(heating_stats, filter_2_heating_list)
            cooling_stats_filter2_list = pool.starmap(cooling_stats, filter_2_cooling_list)

        # Interleave the lists (all_heating, all_cooling, very-cold_heating...)
        filter_1_list = list(chain(*zip(heating_stats_filter1_list, cooling_stats_filter1_list)))
        filter_2_list = list(chain(*zip(heating_stats_filter2_list, cooling_stats_filter2_list)))


    if advanced_filtering:
        stats = list(chain.from_iterable([
                no_filter_list,
                filter_1_list,
                filter_2_list,
                filter_3_list]))
    else:
        stats = list(chain.from_iterable([
                no_filter_list,
                filter_3_list]))

    # Remove empty lists that might have sneaked in
    stats = [stat for stat in stats if stat]

    for stat in stats:
        if stat:
            stats_dict.update({stat['label']: stat})

    def _compute_national_weightings(stats_by_climate_zone, keys, weights):
        def _national_weight(key):
            results = []
            for cz, weight in weights.items():
                stat_cz = stats_by_climate_zone.get(cz)
                if stat_cz is None:
                    value = None
                else:
                    value = stat_cz.get(key)
                if pd.notnull(weight) and pd.notnull(value):
                    results.append((weight, value))
            if len(results) == 0:
                return None
            else:
                weighted_sum = sum([weight * value for weight, value in results])  # noqa: F812
                sum_of_weights = sum([weight for weight, _ in results])
                return weighted_sum / sum_of_weights

        stats = [
            "mean",
            "q1",
            "q2.5",
            "q5",
            "q10",
            "q15",
            "q20",
            "q25",
            "q30",
            "q35",
            "q40",
            "q45",
            "q50",
            "q55",
            "q60",
            "q65",
            "q70",
            "q75",
            "q80",
            "q85",
            "q90",
            "q95",
            "q98",
            "q99",
        ]

        key_stats = [
            "{}_{}".format(key, stat)
            for key in keys for stat in stats
        ]

        return {
            "{}_{}".format(key_stat, "national_weighted_mean"): _national_weight(key_stat)
            for key_stat in key_stats
        }

    def _compute_national_weighting_lower_and_upper_bounds(
            stats_by_climate_zone, keys, weights):

        def _compute_bounds(key):

            # compute sem savings
            means, sems, weights_ = [], [], []
            for cz, weight in weights.items():
                stat_cz = stats_by_climate_zone.get(cz)
                if stat_cz is None:
                    mean, sem = None, None
                else:
                    mean = stat_cz.get("{}_mean".format(key), None)
                    sem = stat_cz.get("{}_sem".format(key), None)

                if pd.notnull(weight) and pd.notnull(mean) and pd.notnull(sem):
                    weights_.append(weight)
                    means.append(mean)
                    sems.append(sem)

            if len(weights_) == 0:
                return {}
            else:
                weighted_sum = sum([
                    weight * mean for weight, mean in zip(weights_, means)  # noqa: F812

                ])
                weighted_mean = weighted_sum / sum(weights_)  # renormalize

                weighted_sem = sum([
                    (weight*sem) ** 2 for weight, sem in zip(weights_, sems)  # noqa: F812

                ]) ** 0.5

                lower_bound = weighted_mean - (1.96 * weighted_sem)
                upper_bound = weighted_mean + (1.96 * weighted_sem)

                return {
                    "{}_lower_bound_95_perc_conf_national_weighted_mean".format(key): lower_bound,
                    "{}_upper_bound_95_perc_conf_national_weighted_mean".format(key): upper_bound
                }

        items = {}
        for key in keys:
            items.update(_compute_bounds(key))
        return items

    national_weighting_stats = []

    if advanced_filtering:
        filters = [
            "no_filter",
            "tau_filter",
            "tau_cvrmse_filter",
            "tau_cvrmse_savings_p01_filter",
        ]
    else:
        filters = [
            "no_filter",
            "tau_cvrmse_savings_p01_filter",
        ]

    climate_zones = [
        "mixed-humid",
        "mixed-dry_hot-dry",
        "marine",
        "hot-humid",
        "very-cold_cold"
    ]
    methods = [
        "baseline_percentile",
        "baseline_regional",
    ]
    for season_type in ["heating", "cooling"]:
        if season_type == "heating":
            weights = HEATING_CLIMATE_ZONE_WEIGHTS
        else:
            weights = COOLING_CLIMATE_ZONE_WEIGHTS

        for filter_ in filters:
            stats_by_climate_zone = {
                cz: stats_dict.get("{}_{}_{}".format(cz, filter_, season_type))
                for cz in climate_zones
            }

            keys = ["percent_savings_{}".format(method) for method in methods]

            national_weightings = _compute_national_weightings(
                stats_by_climate_zone, keys, weights)

            bounds = _compute_national_weighting_lower_and_upper_bounds(
                stats_by_climate_zone, keys, weights)
            national_weightings.update(bounds)

            national_weightings.update(
                {"label": "national_weighted_mean_{}_{}".format(season_type, filter_)}
            )

            national_weighting_stats.append(national_weightings)

    stats = national_weighting_stats + stats
    return stats


def summary_statistics_to_csv(stats, filepath, product_id):
    """ Write metric statistics to CSV file.

    Parameters
    ----------
    stats : list of dict
        List of outputs from thermostat.stats.compute_summary_statistics()
    filepath : str
        Filepath at which to save the suppary statistics
    product_id : str
        A combination of the connected thermostat service plus one or more
        connected thermostat device models that comprises the data set.

    Returns
    -------
    df : pandas.DataFrame
        A pandas dataframe containing the output data.

    """

    if stats is None:
        warnings.warn("No summary statistics to export.")
        return None


    drop_columns = [
        'national_weighted_mean_heating_no_filter',
        'national_weighted_mean_heating_tau_cvrmse_savings_p01_filter',
        'national_weighted_mean_cooling_no_filter',
        'national_weighted_mean_cooling_tau_cvrmse_savings_p01_filter',
        'national_weighted_mean_heating_tau_filter',
        'national_weighted_mean_heating_tau_cvrmse_filter',
        'national_weighted_mean_cooling_tau_filter',
        'national_weighted_mean_cooling_tau_cvrmse_filter',
        ]
    columns = [
        "label",
        "product_id",
        "sw_version",
    ]

    columns.extend([
        "n_thermostat_core_day_sets_total",
        "n_thermostat_core_day_sets_kept",
        "n_thermostat_core_day_sets_discarded",
    ])
    for column_name in REAL_OR_INTEGER_VALUED_COLUMNS_ALL:
        columns.append("{}_n".format(column_name))
        columns.append("{}_upper_bound_95_perc_conf".format(column_name))
        columns.append("{}_mean".format(column_name))
        columns.append("{}_lower_bound_95_perc_conf".format(column_name))
        columns.append("{}_sem".format(column_name))
        for quantile in QUANTILE:
            columns.append("{}_q{}".format(column_name, quantile))

    # add product_id
    for row in stats:
        row["product_id"] = product_id

    # transpose for readability.
    stats_dataframe = pd.DataFrame(stats, columns=columns).set_index('label').transpose()
    # Remove certification columns that are no longer in the stats file
    stats_dataframe.drop(
        drop_columns,
        axis=1,
        errors='ignore',
        inplace=True)
    stats_dataframe.to_csv(filepath)
    return stats_dataframe



import pytest
from thermostat.equipment_type import (
    has_heating,
    has_cooling,
    has_auxiliary,
    has_emergency,
    has_two_stage_heating,
    has_two_stage_cooling,
    has_resistance_heat,
    validate_heat_type,
    validate_cool_type,
    validate_heat_stage,
    validate_cool_stage,
    first_stage_capacity_ratio,
    )



def test_missing_or_blank_heat_type():
    assert(has_heating(None) is False)
    assert(has_heating('') is False)


def test_missing_or_blank_cool_type():
    assert(has_cooling(None) is False)
    assert(has_cooling('') is False)


def test_missing_or_blank_heat_stage():
    assert(has_two_stage_heating(None) is False)
    assert(has_two_stage_heating('') is False)


def test_missing_or_blank_cool_stage():
    assert(has_two_stage_cooling(None) is False)
    assert(has_two_stage_cooling('') is False)


def test_has_heating():
    for i in [
        'furnace_or_boiler',  # Non heat pump heating (gas or oil furnace, electric resistance)
        'heat_pump_electric_backup',   # Heat pump with electric resistance heat (strip heat)
        'heat_pump_no_electric_backup',  # Heat pump without electric resistance heat
    ]:
        assert(has_heating(i) is True)
    for i in [
        'heat_pump_dual_fuel',  # Dual fuel heat pump (e.g. gas or oil fired)
        'other',  # Multi-zone, ?
        'none',  # None
    ]:
        assert(has_heating(i) is False)
    assert(has_heating(None) is False)
    assert(validate_heat_type('furnace_or_boiler') is True)
    assert(validate_heat_type('bogus_heat_pump') is False)


def test_has_cooling():
    for i in [
        'heat_pump',  # Heat pump w/ cooling
        'central',  # Central AC
    ]:
        assert(has_cooling(i) is True)
    assert(has_cooling('none') is False)
    assert(has_cooling('other') is False)
    assert(has_cooling(None) is False)
    assert(validate_cool_type('heat_pump') is True)
    assert(validate_cool_type('bogus_pump') is False)


def test_has_two_stage_cooling():
    assert(has_two_stage_cooling('two_speed') is True)
    assert(has_two_stage_cooling('two_stage') is True)
    assert(has_two_stage_cooling('single_speed') is False)
    assert(has_two_stage_cooling('single_stage') is False)
    assert(has_two_stage_cooling('modulating') is False)
    assert(has_two_stage_cooling(None) is False)
    assert(has_two_stage_cooling('none') is False)
    assert(has_two_stage_cooling('') is False)
    assert(validate_cool_stage('two_speed') is True)
    assert(validate_cool_stage('ten_speed') is False)


def test_has_two_stage_heating():
    assert(has_two_stage_heating('two_stage') is True)
    assert(has_two_stage_heating('single_stage') is False)
    assert(has_two_stage_heating('two_speed') is True)
    assert(has_two_stage_heating('single_speed') is False)
    assert(has_two_stage_heating('modulating') is False)
    assert(has_two_stage_heating(None) is False)
    assert(has_two_stage_heating('none') is False)
    assert(has_two_stage_heating('') is False)
    assert(validate_heat_stage('two_stage') is True)
    assert(validate_heat_stage('two_speed') is True)
    assert(validate_heat_stage('2_stage') is False)


def test_has_emergency():
    assert(has_emergency('heat_pump_electric_backup') is True)
    assert(has_emergency('heat_pump_no_electric_backup') is False)
    assert(has_emergency('heat_pump_dual_fuel') is False)


def test_has_auxiliary():
    assert(has_auxiliary('heat_pump_electric_backup') is True)
    assert(has_auxiliary('heat_pump_no_electric_backup') is False)
    assert(has_auxiliary('heat_pump_dual_fuel') is False)


def test_has_resistance_heat():
    assert(has_resistance_heat('heat_pump_electric_backup') is True)
    assert(has_resistance_heat('heat_pump_no_electric_backup') is False)
    assert(has_resistance_heat('heat_pump_dual_fuel') is False)


def test_ratio():
    assert(first_stage_capacity_ratio('furnace_or_boiler') == .65)
    assert(first_stage_capacity_ratio('heat_pump') == .72)
